import os 
import sys
import time
import re
import platform
import threading
import posixpath
import ast
import textwrap
import binascii
import shutil
import stat
import subprocess
import json
import urllib.request
import pathlib
from typing import Callable, Optional, Tuple, Union
import click

import serial
from serial.tools import list_ports
from genlib.ansiec import ANSIEC
 

from . import __version__  

#--------------------------------------------------------------

_stdout_lock = threading.Lock()
_buffer = b''
_expected_bytes = 0

_error_header       = b"Traceback (most recent call last):"
_error_header_buf   = b""
_skip_error_output  = False

def stdout_write_bytes(b):
    global _buffer, _expected_bytes
    global _error_header_buf, _skip_error_output

    if _skip_error_output:
        return

    _error_header_buf = (_error_header_buf + b)[-len(_error_header):]
    if _error_header_buf == _error_header:
        _skip_error_output = True
        return
    
    if b in (b'\x04', b''): # EOF or empty byte
        return

    with _stdout_lock:
        if _expected_bytes:
            _buffer += b
            _expected_bytes -= 1

            if _expected_bytes == 0: 
                sys.stdout.buffer.write(_buffer)
                sys.stdout.buffer.flush()
                _buffer = b''
                return 

        elif b[0] <= 0x7F:              # ASCII
            sys.stdout.buffer.write(b)
            sys.stdout.buffer.flush()
        else:                           # UTF-8 
            hdr = b[0]   
            if   hdr & 0xF8 == 0xF0: _expected_bytes = 3  
            elif hdr & 0xF0 == 0xE0: _expected_bytes = 2  
            elif hdr & 0xE0 == 0xC0: _expected_bytes = 1
            else:                       # Invalid UTF-8   
                sys.stdout.buffer.write(b.hex().encode())
                return 
            _buffer = b                  # save first byte

#--------------------------------------------------------------

IS_WINDOWS: bool = platform.system() == "Windows"
CR, LF = b"\r", b"\n"

_EXTMAP : dict[str, bytes] = {
    "H": b"\x1b[A",   # ↑
    "P": b"\x1b[B",   # ↓
    "M": b"\x1b[C",   # →
    "K": b"\x1b[D",   # ←
    "G": b"\x1b[H",   # Home
    "O": b"\x1b[F",   # End
    "R": b"\x1b[2~",  # Ins
    "S": b"\x1b[3~",  # Del
}

def _utf8_need_follow(b0: int) -> int:
    if b0 & 0b1000_0000 == 0:          # 0xxxxxxx → ASCII
        return 0
    if b0 & 0b1110_0000 == 0b1100_0000:    # 110xxxxx
        return 1
    if b0 & 0b1111_0000 == 0b1110_0000:    # 1110xxxx
        return 2
    if b0 & 0b1111_1000 == 0b1111_0000:    # 11110xxx
        return 3
    return 0                               

if IS_WINDOWS:
    import msvcrt
    from typing import Callable

    def getch() -> bytes:
        w = msvcrt.getwch()
        if w in ("\x00", "\xe0"):  # arrow keys ect.
            return _EXTMAP.get(msvcrt.getwch(), b"")
        return w.encode("utf-8")
else:                            
    import tty
    import termios

    _FD, _OLD = sys.stdin.fileno(), termios.tcgetattr(sys.stdin)

    def _raw(on: bool):
        tty.setraw(_FD) if on else termios.tcsetattr(_FD, termios.TCSADRAIN, _OLD)

    def getch() -> bytes:
        try:
            _raw(True)
            first = os.read(_FD, 1)
            need = _utf8_need_follow(first[0])
            return first + (os.read(_FD, need) if need else b"")
        finally:
             _raw(False)

if IS_WINDOWS:
    import msvcrt
    from typing import Callable

    _PUTB: Callable[[bytes], None] = msvcrt.putch 
    _PUTW: Callable[[str], None]   = msvcrt.putwch 
    
    def _write_bytes(data: bytes) -> None:
        sys.stdout.buffer.write(data)
        sys.stdout.flush()
        
    def putch(data: bytes) -> None:
        if data == CR: 
            _PUTB(LF);  return

        if len(data) > 1 and data.startswith(b"\x1b["):
            _write_bytes(data)   
        elif len(data) == 1 and data < b"\x80":  
            _PUTB(data)
        else:                                         
            _PUTW(data.decode("utf-8", "strict"))  
          
else:
    def putch(data: bytes) -> None:
        if data != CR:
            sys.stdout.buffer.write(data)
            sys.stdout.flush()

#--------------------------------------------------------------
class UpyBoardError(BaseException):
    """
    Custom exception for UpyBoard operations.
    """
    def __init__(self, message: str):
        super().__init__(message)
        self.message = message

    def __str__(self):
        return f"UpyBoardError: {self.message}"


class UpyBoard:
    """
    A class to interact with a MicroPython device over a serial connection.
    This class provides methods to execute commands, manage files, and interact with the device's REPL.
    """
    _REPL_BUFSIZE = 2048  # Min 256  
    
    _CTRL_A = b'\x01'
    _CTRL_B = b'\x02'
    _CTRL_C = b'\x03'
    _CTRL_D = b'\x04'
    _EOF_MARKER = b'\x04'
    _RAW_REPL_PROMPT = b'raw REPL; CTRL-B to exit\r\n>'
    _SOFT_REBOOT_MSG = b'soft reboot\r\n'
    _OK_RESPONSE = b'OK'
    _DEIVCE_CHUNK_SIZES = 4096
    
    def __init__(self, port:str, baudrate:int=115200):
        """
        Initialize the UpyBoard instance with the specified serial port and baud rate.
        :param port: The serial port to connect to.
        :param baudrate: The baud rate for the serial connection (default is 115200).
        :raises UpyBoardError: If the serial port cannot be opened or if the device is not found.
        """
        try:
            self.serial = serial.Serial(port, baudrate, timeout=1.0, write_timeout=1.0)
            self.__serial_lock = threading.Lock()
                            
        except serial.SerialException as e:
            raise UpyBoardError(f"failed to open {port} ({e})")
        except (OSError, IOError): 
            raise UpyBoardError(f"failed to open {port} (device not found)")
    
        self.__init_repl()

    def __init_repl(self):
        """
        Initialize the REPL (Read-Eval-Print Loop) for the UpyBoard.
        This function sets up the serial connection and prepares the board for REPL interaction.
        """
        self.serial_reader_running = None
        self.serial_out_put_enable = True
        self.serial_out_put_count = 0

    def __write(self, data:bytes) -> None:
        """
        Write data to the serial port.
        :param data: The data to write to the serial port.
        """
        self.serial.write(data)

    def __read(self, n:int=1) -> bytes:
        """
        Read a specified number of bytes from the serial port.
        :param n: Number of bytes to read from the serial port.
        :return: The bytes read from the serial port.
        """
        return self.serial.read(n)
    
    def __read_ex(self, min_num_bytes:int, ending:bytes, timeout:int=5, data_consumer:Optional[Callable[[bytes], None]]=None) -> bytes:
        """
        Read data from the serial port until the specified ending is found or timeout occurs.
        :param min_num_bytes: Minimum number of bytes to read before checking for the ending.
        :param ending: The byte sequence that indicates the end of the data.
        :param timeout: Maximum time to wait for the ending in seconds (0 = no timeout).
        :param data_consumer: Optional callback function to process the data as it is read.
        :return: The data read from the serial port as bytes.
        :raises UpyBoardError: If timeout occurs or serial communication fails.
        """
        data = b''
        start_time = time.time()
        last_activity = start_time
        
        try:
            if min_num_bytes > 0:
                initial_data = self.__read(min_num_bytes)
                if not initial_data:
                    pass
                else:
                    data += initial_data
                    if data_consumer:
                        data_consumer(initial_data)
                    last_activity = time.time()
            
            while not data.endswith(ending):
                current_time = time.time()
                
                if timeout > 0 and (current_time - start_time) >= timeout:
                    break
                
                if self.serial.in_waiting > 0:
                    new_data = self.__read(1)
                    if new_data:
                        data += new_data
                        if data_consumer:
                            data_consumer(new_data)
                        last_activity = current_time
                else:
                    time.sleep(0.01)                    
                    if timeout > 0 and (current_time - last_activity) > max(timeout * 2, 10):
                        break
                    
            return data
            
        except serial.SerialException as e:
            raise UpyBoardError(f"Serial communication error: {e}")
        except Exception as e:
            if isinstance(e, UpyBoardError):
                raise
            return data

    def __enter_repl(self, soft_reset:bool=True):
        """
        Enter the raw REPL mode of the device.
        This function sends the necessary commands to the device to enter the raw REPL mode.
        :param soft_reset: If True, perform a soft reset before entering the raw REPL.
        """
        # Clear any pending data
        self.serial.write(b'\r' + self._CTRL_C + self._CTRL_C)  # Double Ctrl-C
        time.sleep(0.1)
        self.serial.reset_input_buffer()
        
        # Enter raw REPL
        self.serial.write(b'\r' + self._CTRL_A)
        
        if soft_reset:
            data = self.__read_ex(1, self._RAW_REPL_PROMPT, timeout=3)
            if not data.endswith(self._RAW_REPL_PROMPT):
                raise UpyBoardError('could not enter raw repl')

            self.serial.write(self._CTRL_D)  # soft reset
            data = self.__read_ex(1, self._SOFT_REBOOT_MSG, timeout=3)
            if not data.endswith(self._SOFT_REBOOT_MSG):
                raise UpyBoardError('could not enter raw repl')
            
        # Wait for the raw REPL prompt
        data = self.__read_ex(1, self._RAW_REPL_PROMPT[:-1], timeout=3)
        if not data.endswith(self._RAW_REPL_PROMPT[:-1]):
            raise UpyBoardError('could not enter raw repl')
            
    def __leave_repl(self):
        """
        Leave the raw REPL mode of the device.
        """
        self.serial.write(b'\r' + self._CTRL_B)  # enter friendly REPL
        
    def __follow_task(self, echo:bool):
        """
        Follow the REPL output and echo input characters.
        This function reads characters from the input and writes them to the serial port.
        If echo is True, it also writes the characters to stdout.
        :param echo: If True, echo the input characters to stdout.
        """
        try:
            while True:
                ch = getch()    
                if ch == self._CTRL_C:
                    break
                if echo:
                    putch(ch)
                self.__write(CR if ch == LF else ch)
        finally:
            if _core == "EFR32MG":
                self.serial.write(self._CTRL_C)  # interrupt a running program
            else:
                self.__reset()
            sys.exit(0)
        
    def __exec(self, command:str|bytes, stream_output:bool=False, echo:bool=False) -> bytes:
        """
        Execute a command on the device and return the output.
        :param command: The command to execute.
        :param stream_output: If True, stream the output to stdout.
        :param echo: If True, echo the command to stdout.
        :return: The output of the command as bytes.
        """
        if isinstance(command, str):
            command = command.encode('utf-8')
            
        data_consumer = stdout_write_bytes if stream_output else None

        data = self.__read_ex(1, b'>')
        if not data.endswith(b'>'):
            raise UpyBoardError('could not enter raw repl')

        command_len = len(command)
        
        current_buffer_size = 1024
        max_buffer_size = 8192
        bytes_sent = 0
        
        # Adaptive sending with performance monitoring
        start_time = time.time()
        
        while bytes_sent < command_len:
            chunk_start = bytes_sent
            chunk_end = min(bytes_sent + current_buffer_size, command_len)
            chunk = command[chunk_start:chunk_end]
            
            chunk_start_time = time.time()
            self.serial.write(chunk)
            bytes_sent += len(chunk)
            
            chunk_time = time.time() - chunk_start_time
            if chunk_time < 0.01 and current_buffer_size < max_buffer_size:
                current_buffer_size = min(current_buffer_size * 2, max_buffer_size)
            elif chunk_time > 0.05:
                current_buffer_size = max(current_buffer_size // 2, 512)
            
            if bytes_sent % 32768 == 0:  # Every 32KB
                time.sleep(0.005)  # 5ms pause
        
        self.serial.write(self._EOF_MARKER)
        
        transfer_time = time.time() - start_time
        timeout = max(5, int(transfer_time * 2))
        
        data = self.__read_ex(1, self._OK_RESPONSE, timeout=timeout)
        if not data.endswith(self._OK_RESPONSE):
            raise UpyBoardError('could not execute command (response: %r)' % data)
        
        if stream_output:
            threading.Thread(target=self.__follow_task, args=(echo,), daemon=True).start()
        
        # Read command output
        data = self.__read_ex(1, self._EOF_MARKER, 0, data_consumer)
        if not data.endswith(self._EOF_MARKER):
            raise UpyBoardError('timeout waiting for first EOF reception')
        data = data[:-1]

        # Read error output
        data_err = self.__read_ex(1, self._EOF_MARKER, 0, data_consumer)
        if not data_err.endswith(self._EOF_MARKER):
            raise UpyBoardError('timeout waiting for second EOF reception')
        data_err = data_err[:-1]

        if data_err:
            raise UpyBoardError(data_err.decode('utf-8', errors='replace'))

        return data
                
    def __drain_eof(self, max_ms:int=200):
        """
        Drain the serial input buffer until EOF is received or timeout occurs.
        :param max_ms: Maximum time to wait for EOF in milliseconds.
        """
        deadline = time.time() + max_ms / 1000
        while time.time() < deadline:
            waiting = self.serial.in_waiting
            if waiting:
                _ = self.__read(waiting) 
            else:
                time.sleep(0.01)
                
    def __repl_serial_to_stdout(self):        
        """
        Read data from the serial port and write it to stdout.
        """
        def hexsend(string_data=''):
            import binascii
            hex_data = binascii.unhexlify(string_data)
            return hex_data

        try:
            data = b''
            try:
                while self.serial_reader_running:
                    count = self.serial.in_waiting
                    if count == 0: 
                        time.sleep(0.01)
                        continue

                    if count > 0:
                        data += self.serial.read(count)

                        if len(data) < 20:
                            try:
                                data.decode()
                            except UnicodeDecodeError:
                                continue

                        if data != b'':
                            if self.serial_out_put_enable and self.serial_out_put_count > 0:
                                if platform.system() == 'Windows':   
                                    sys.stdout.buffer.write(data.replace(b"\r", b""))
                                else:
                                    sys.stdout.buffer.write(data)
                                    
                                sys.stdout.buffer.flush()
                        else:
                            self.serial.write(hexsend(data))

                        data = b''
                        self.serial_out_put_count += 1
            except:
                print('')
                return
        except KeyboardInterrupt:
            if serial != None:
                serial.close()

    def __fs_ls_fallback(self, dir:str="/") -> list[Tuple[str, int, bool]]:
        """
        Fallback method for listing directory contents (original implementation).
        """
        if not dir.startswith("/"):
            dir = "/" + dir
            
        command = f"""
            import os
            def listdir(dir):
                if dir == '/':                
                    return sorted([dir + f for f in os.listdir(dir)])
                else:
                    return sorted([dir + '/' + f for f in os.listdir(dir)])
            print(listdir('{dir}'))
        """
        out = self.exec(command)
        file_list = ast.literal_eval(out.decode("utf-8"))
        
        # Convert to detailed format for compatibility
        result = []
        for f in file_list:
            f_name = f.split("/")[-1]
            is_dir = self.fs_is_dir(f)
            size = 0 if is_dir else self.fs_state(f)
            result.append([f_name, size, is_dir])
        
        return result

    def __reset(self):
        """
        Reset the device by executing a soft reset command. 
        """
        command = f"""
            import machine
            machine.soft_reset()  # Ctrl+D
        """
        self.exec(command)
        
    def exec(self, command:str=None):
        """
        Run a command or script on the device.
        :param command: The command to execute.
        """
        self.__enter_repl()
        try:
            command = textwrap.dedent(command)
            out = self.__exec(command)
        finally:
            self.__leave_repl()
    
        return out
    
    def run(self, local, stream_output:bool=False, echo:bool=False):
        """
        Run a command or script on the device.
        :param local: Path to the script file to execute.
        :param stream_output: If True, stream the output to stdout.
        :param echo: If True, echo the command to stdout.
        """
        self.__enter_repl()
        try:
            with open(local, "rb") as f:
                data = f.read()
            self.__exec(data, stream_output, echo)
            self.__drain_eof(max_ms=200)
        finally:
            self.__leave_repl()
    
    def close(self):
        """
        Close the serial connection.
        """
        self.serial.close()

    def reset(self):
        self.__write(b'\r' + self._CTRL_D)  
 
    def repl(self):
        """
        Enter the REPL mode, allowing interaction with the device.
        """
        self.serial_reader_running = True
        self.serial_out_put_enable = True
        self.serial_out_put_count = 1

        self.__reset()
        self.__read_ex(1, b'\x3E\x3E\x3E', timeout=1) # read prompt >>>

        repl_thread = threading.Thread(target=self.__repl_serial_to_stdout, daemon=True, name='REPL')
        repl_thread.start()

        self.serial.write(b'\r') # Update prompt
        
        while True:
            char = getch()

            if char == b'\x07': 
                self.serial_out_put_enable = False
                continue
            elif char == b'\x0F': 
                self.serial_out_put_enable = True
                self.serial_out_put_count = 0
                continue
            elif char == b'\x00' or not char: # Ignore null characters
                continue
            elif char == self._CTRL_C:  # Ctrl + C to exit repl mode
                break
            
            try:
                self.serial.write(b'\r' if char == b'\n' else char)
            except:
                break
            
        self.serial_reader_running = False
        self.__reset()
        print('')

    def fs_get(self, remote:str, local:str):
        """ 
        Download a file from the connected device to the local filesystem.
        :param remote: The path to the file on the device.
        :param local: The local path where the file should be saved.
        :raises UpyBoardError: If the file is empty or if the download fails.
        """
        local_file = None
        
        if local:
            if os.path.isdir(local):
                local = os.path.join(local, os.path.basename(remote))
            local_file = open(local, "wb")
        
        bytes_read = 0
        bar_length = 40 

        if local_file:
            print(f"{ANSIEC.FG.BRIGHT_BLUE}{remote.replace(_device_root_fs, '', 1)}{ANSIEC.OP.RESET}")

        try:
            file_size = self.fs_state(remote)

            self.__enter_repl()

            init_command = f"""
                import sys
                f = open('{remote}', 'rb')
                """
            self.__exec(textwrap.dedent(init_command))
                            
            while bytes_read < file_size:
                remaining = min(self._DEIVCE_CHUNK_SIZES, file_size - bytes_read)

                read_cmd = f"""
                    chunk = f.read({remaining})
                    if chunk:
                        sys.stdout.buffer.write(chunk)
                    """
                chunk_data = self.__exec(textwrap.dedent(read_cmd))
                
                if chunk_data:
                    if local_file:
                        local_file.write(chunk_data)
                    else:
                         sys.stdout.buffer.write(chunk_data)
                         sys.stdout.flush()
                
                    bytes_read += len(chunk_data)
                                        
                    if local_file:
                        pct = bytes_read / file_size
                        block = int(round(bar_length * pct))
                        bar = "#" * block + "-" * (bar_length - block)
                        percent = int(pct * 100)
                        print(f"{ANSIEC.OP.left()}[{bar}] {percent}% ({bytes_read}/{file_size})", end="", flush=True)
                else:
                    break

            self.__exec("f.close()")
        
        except Exception as e:
            raise UpyBoardError(f"Download failed: {e}")
        finally:
            self.__leave_repl()
            if local_file:
                print()
                local_file.close()

        if bytes_read != file_size:
            raise UpyBoardError(f"Download incomplete: got {bytes_read}/{file_size} bytes")
    
    def fs_state(self, path:str) -> int:
        """
        Return file size of given path.
        """
        if _core == "EFR32MG":
            command = f"""
                try:
                    with open('{path}', 'rb') as f:
                        f.seek(0, 2)
                        size = f.tell()
                    print(size)
                except Exception as e:
                    print(0)
            """
            out = self.exec(command)
            return int(out.decode('utf-8'))
        else:
            command = f"""
                import os
                try:
                    st = os.stat('{path}')
                    print(st[6])
                except:
                    print(0)
            """
        out = self.exec(command)
        return int(out.decode('utf-8'))

    def fs_is_dir(self, path:str) -> bool:
        """
        Check if the given path is a directory.
        :param path: The path to check.
        :return: True if the path is a directory, False otherwise.
        """
        command = f"""
            vstat = None
            try:
                from os import stat
            except ImportError:
                from os import listdir
                vstat = listdir
            def ls_dir(path):
                if vstat is None:
                    return stat(path)[0] & 0x4000 != 0
                else:
                    try:
                        vstat(path)
                        return True
                    except OSError as e:
                        return False
            print(ls_dir('{path}'))
        """
        out = self.exec(command)
        return ast.literal_eval(out.decode("utf-8"))

    def fs_ls_detailed(self, dir:str="/") -> list[Tuple[str, int, bool]]:
        """
        List the contents of a directory with detailed information (size, type) in a single operation.
        :param dir: The directory to list. Defaults to the root directory ("/").
        :return: A list of tuples containing (name, size, is_dir) for each item.
        """
        if not dir.startswith("/"):
            dir = "/" + dir

        command = f"""
            import os
            import json
            import sys
            def xbee3_zigbee_state(path):
                try:
                    with open(path, 'rb') as f:
                        f.seek(0, 2)
                        size = f.tell()
                    return size
                except Exception as e:
                    return 0

            def get_detailed_listing(path):
                try:
                    items = []
                    for item in os.listdir(path):
                        full_path = path + ('/' + item if path != '/' else item)
                        if sys.platform == 'xbee3-zigbee':
                            is_dir = False
                            size = xbee3_zigbee_state(full_path)
                            if size == 0:
                                is_dir = True
                            items.append([item, size, is_dir])
                            continue
                        try:
                            stat_info = os.stat(full_path)
                            is_dir = stat_info[0] & 0x4000 != 0
                            size = 0 if is_dir else stat_info[6]
                            items.append([item, size, is_dir])
                        except:
                            # If stat fails, try to determine if it's a directory
                            try:
                                os.listdir(full_path)
                                items.append([item, 0, True])  # It's a directory
                            except:
                                items.append([item, 0, False])  # It's a file
                    return sorted(items, key=lambda x: (not x[2], x[0].lower()))
                except Exception as e:
                    return []

            print(json.dumps(get_detailed_listing('{dir}')))
        """

        try:
            out = self.exec(command)
            result = json.loads(out.decode("utf-8").strip())
            return result
        except (json.JSONDecodeError, UpyBoardError):
            return self.__fs_ls_fallback(dir)

    def fs_mkdir(self, dir:str) -> bool:       
        """
        Create a directory on the connected device.
        :param dir: The directory to create.
        :return: True if the directory was created, False if it already exists.
        """
        command = f"""
            import os
            def mkdir(dir):
                parts = dir.split(os.sep)
                dirs = [os.sep.join(parts[:i+1]) for i in range(len(parts))]
                check = 0
                for d in dirs:
                    try:
                        os.mkdir(d)
                    except OSError as e:
                        check += 1
                        if "EEXIST" in str(e):
                            continue
                        else:
                            return False
                return check < len(parts)
            print(mkdir('{dir}'))
        """        
        out = self.exec(command)

        return ast.literal_eval(out.decode("utf-8"))

    def fs_putdir(self, local:str, remote:str):
        """
        Upload a directory and its contents to the connected device.
        :param local: The local directory to upload.
        :param remote: The remote directory path on the device.
        :param callback: Optional callback function to call with the remote file path and size.
        """        
        for parent, child_dirs, child_files in os.walk(local, followlinks=True):
            remote_parent = posixpath.normpath(posixpath.join(remote, os.path.relpath(parent, local)))
           
            try:
                self.fs_mkdir(remote_parent)
            except:
                pass

            for filename in child_files:
                local = os.path.join(parent, filename)
                remote = posixpath.join(remote_parent, filename)      
                self.fs_put(local, remote)

    def fs_put(self, local:str, remote:str):
        """
        Upload a file to the connected device.
        :param local: The local file path to upload.
        :param remote: The remote file path on the device.
        :raises UpyBoardError: If the upload fails or if the file already exists.
        """
        sent = 0
        bar_length = 40
        total = os.path.getsize(local)
         
        self.__enter_repl()
        try:
            self.__exec(f"f = open('{remote}', 'wb')")                
        except UpyBoardError as e:
            if "EEXIST" in str(e):
                self.__leave_repl()
                self.fs_rm(remote)
                self.fs_put(local, remote)
                return

        try:    
            f = open(local, "rb")
            print(f"{ANSIEC.FG.BRIGHT_BLUE}{remote.replace(_device_root_fs, '', 1)}{ANSIEC.OP.RESET}")

            while True:
                chunk = f.read(self._DEIVCE_CHUNK_SIZES)
                if not chunk:
                    break
                
                self.__exec(f"f.write({repr(chunk)})")
    
                sent += len(chunk)
                pct = sent / total
                block = int(round(bar_length * pct))
                bar = "#" * block + "-" * (bar_length - block)
                percent = int(pct * 100)
                print(f"{ANSIEC.OP.left()}[{bar}] {percent}% ({sent}/{total})", end="", flush=True)
            print()
                    
            self.__exec("f.close()")
        finally:
            self.__leave_repl()

    def fs_rm(self, filename:str):
        """
        Remove a file from the connected device.
        :param filename: The file to remove.
        """
        command = f"""
            import os
            os.remove('{filename}')
        """
        self.exec(command)

    def fs_rmdir(self, dir:str):
        """
        Remove a directory and all its contents recursively.
        :param dir: The directory to remove.
        """
        command = f"""
            import os
            def rmdir(dir):
                os.chdir(dir)
                for f in os.listdir():
                    try:
                        os.remove(f)
                    except OSError:
                        pass
                for f in os.listdir():
                    rmdir(f)
                os.chdir('..')
                os.rmdir(dir)
            rmdir('{dir}')
        """
        self.exec(command)

    def fs_format(self) -> bool:
        """
        Format the filesystem of the connected device based on its core type.
        :return: True if the filesystem was successfully formatted, False otherwise.
        """
        if _core == "ESP32":
            command = """ 
                import os
                os.fsformat('/flash')
            """
        elif _core in ("ESP32S3", "ESP32C6"):
            command = """
                import os
                from flashbdev import bdev
                os.umount('/')
                os.VfsLfs2.mkfs(bdev)
                os.mount(bdev, '/')
            """
        elif _core == "EFR32MG":
            command = """
                import os
                os.format()
            """
        elif _core == "RP2350":
            command = """
                import os
                import rp2
                bdev = rp2.Flash()
                os.VfsFat.mkfs(bdev)
            """
        else:
            return False

        try:
            self.exec(command)
        except UpyBoardError:
            return False
            
        return True
    
    def fs_df(self):
        """
        Get filesystem information including total, used, free space and usage percentage.
        :return: A tuple containing total space, used space, free space, and usage percentage.
        """
        command = f"""
            import os
            import json
            def get_fs_info(path='/'):
                stats = os.statvfs(path)
                block_size = stats[0]
                total_blocks = stats[2]
                free_blocks = stats[3]

                total = block_size * total_blocks
                free = block_size * free_blocks
                used = total - free
                usage_pct = round(used / total * 100, 2)
                
                return total, used, free, usage_pct
            print(get_fs_info())
        """
        out = self.exec(command)
        return ast.literal_eval(out.decode("utf-8"))

#--------------------------------------------------------------

def load_env_from_upyboard():
    """
    Load environment variables from the .upyboard file in the .vscode directory.
    This function searches for the .upyboard file in the current directory and its parent directories,
    and loads the key-value pairs into the environment variables.
    """
    current_path = os.getcwd()

    while True:
        upyboard_path = os.path.join(current_path, ".vscode", ".upyboard")
        if os.path.isfile(upyboard_path):
            with open(upyboard_path) as f:
                for line in f:
                    if '=' in line and not line.strip().startswith('#'):
                        key, val = line.strip().split('=', 1)
                        os.environ[key.strip()] = val.strip()
            break
        
        parent_path = os.path.dirname(current_path)
        if parent_path == current_path:
            break
        current_path = parent_path


class ColorfulGroup(click.Group):
    """
    Custom Click group that formats commands with colors and styles.
    """
    def format_commands(self, ctx, formatter):
        """
        Format the commands in the group with colors and styles.
        :param ctx: Click context object.
        :param formatter: Formatter object to format the output.
        """
        commands = self.list_commands(ctx)
        rows = []
        for cmd_name in commands:
            cmd = self.get_command(ctx, cmd_name)
            if cmd is None or cmd.hidden:
                continue
            help_text = cmd.get_short_help_str()
            cmd_display = click.style(cmd_name, fg='green', bold=True)
            rows.append((cmd_display, help_text))
        if rows:
            with formatter.section('Commands'):
                formatter.write_dl(rows)


SUPPORT_CORE_DEVICE_TYPES = {  #  core, platform, device
    'EFR32MG':{'zigbee':'xnode'}, 
    'ESP32':{'lopy4':'smartfarm1'}, 
    'ESP32S3':{},
    'ESP32C6':{},
    'RP2350':{
        'pico2w0':'ticle', 
        'pico2w1':'xconvey', 
        'pico2w2':'xhome', 
        'pico2w3':'autocon'
        }, 
    'RP3A0':{
        'pi02w0':'ticle_ex',
        }
    }

def get_micropython_board_info(port:str, is_long:bool=False) -> Optional[Tuple[str, str, str, str]]:
    """
    Get the firmware version, build date, core name, and device name of the connected device.
    :param port: The port name of the connected device.
    :param is_long: If True, display detailed information about the connected device.
    :return: The (version, date, core, device) or string of the connected device.
    """
    try:
        with serial.Serial(port, 115200, timeout=1) as ser:
            ser.write(b'\r\x03') 
            time.sleep(0.1)
            ser.reset_input_buffer()

            ser.write(b'\r\x02')
            time.sleep(0.1)

            response = ser.read_all().decode(errors='ignore').strip()
            if response:
                m = re.search(r"(?:MicroPython|Pycom MicroPython)\s+(.*)", response)
                if m:
                    response = m.group(1)

                if is_long:
                    return response

                rx = re.compile(
                    r"(?P<full_version>[^\s\[,]+),?"
                    r"(?:\s*\[[^\]]+\])?"
                    r"\s+on\s+(?P<date>\d{4}-\d{2}-\d{2});\s+"
                    r"(?P<manufacturer>.+?)\s+with\s+(?P<core>\S+)",
                    re.I,
                )
                m = rx.search(response)
                if not m:
                    return None

                full_version = m.group("full_version").lstrip("v").rstrip(",")
                date         = m.group("date")
                manufacturer = m.group("manufacturer").strip().lower()
                core         = m.group("core").strip().upper() 
                
                if len(manufacturer.split()) > 1:
                    manufacturer = manufacturer.split()[-1]
                    
                if manufacturer.startswith('pico2'):
                    manufacturer = manufacturer[:-1]
                
                num_match = re.match(r"(\d+\.\d+)", full_version)
                pico_match = re.match(r"pico2_w_(\d{4})_(\d{2})_(\d{2})", full_version, re.I)  # only pimoroni -> pico2_w_2024_01_01

                if num_match:
                    version = num_match.group(1)
                elif pico_match:
                    y, mth, d = pico_match.groups()

                    if int(y) >= 2025:
                        version = 1.25
                    else:
                        version = 1.24
                else:
                    version = "?"
                    
                device_list = SUPPORT_CORE_DEVICE_TYPES.get(core, None)
                if device_list:
                    device = device_list.get(manufacturer, core)
                else:
                    device = core

                return version, date, core, device
    except (OSError, serial.SerialException):
        pass
    
    return None


#--------------------------------------------------------------------------------------

import upyboard

def __tiny_command(cmd:str) -> None:
    """
    Execute a command on the connected device, wrapping it if necessary.
    :param cmd: The command to execute.
    """
    import ast

    try:
        tree = ast.parse(cmd, mode="exec")
        is_expr = (
            len(tree.body) == 1 and
            isinstance(tree.body[0], ast.Expr)
        )
    except SyntaxError:
        is_expr = False

    if is_expr:
        wrapped = (
            f"__r={cmd}\n"
            "if __r is not None:\n"
            "    print(repr(__r))\n"
        )
    else:
        wrapped = cmd if cmd.endswith("\n") else cmd + "\n"

    out = _upy.exec(wrapped)
    print(out.decode("utf-8", "replace"), end="", flush=True)

_upy = None
_version = 0.0
_core = ""
_device = ""
_device_root_fs = "/"
_core_path = ""
_device_path = ""
_sport = ""

@click.group(cls=ColorfulGroup, invoke_without_command=True)
@click.option(
    "--sport",
    "-s",
    envvar="SERIAL_PORT",
    default="",
    type=click.STRING,
    help="The serial port name for connected device.",
    metavar="SPORT",
)
@click.option(
    "--baud",
    '-b',
    envvar="SERIAL_BAUD",
    default=115200,
    type=click.INT,
    help="Baud rate set only when the connected device uses UART. (Default 115200)",
    metavar="BAUD",
)
@click.option(
    "--command",
    '-c',
    default="",
    type=click.STRING,
    help=" Command to execute on the connected device.",
)   
@click.version_option(__version__, "-v", "--version", message="upyboard %(version)s")
@click.pass_context
def cli(ctx, sport, baud, command):
    """
    UpyBoard CLI - A command line interface for managing MicroPython devices.
    
    This tool allows you to interact with MicroPython devices connected via serial port.
    It supports various commands to manage files, execute scripts, and interact with the device's filesystem.
    """
    global _upy, _version, _core, _device, _device_root_fs, _core_path, _device_path, _sport

    _sport = sport
    
      
    if ctx.invoked_subcommand in ("scan", "sport"):
            return

    descript =  get_micropython_board_info(_sport)
    if not descript:
        if _sport:
            print(f"There is no device connected to the {ANSIEC.FG.BRIGHT_RED}{_sport}{ANSIEC.OP.RESET}.")
        else:
            print(f"The serial port name is missing.")
        raise click.Abort()

    _version, _, _core, _device = descript
    _version = float(_version)
    
    if not _core in (SUPPORT_CORE_DEVICE_TYPES.keys()):
        print(f"The {ANSIEC.FG.BRIGHT_RED}{_device}{ANSIEC.OP.RESET} is not supported.")
        raise click.Abort()
    
    if _device in ('xnode', 'smartfarm1'):
        _device_root_fs = "/flash/"
           
    _core_path = os.path.join(os.path.dirname(upyboard.__file__), os.path.join("core", _core))
    if _core != _device:
        _device_path = os.path.join(os.path.dirname(upyboard.__file__), os.path.join("device", _device))  
        
    try:
        _upy = UpyBoard(_sport, baud)
    except UpyBoardError:
        print(f"Device is not connected to {ANSIEC.FG.BRIGHT_RED}{_sport}{ANSIEC.OP.RESET}")
        print(f"Please check the port with the scan command and try again.")
        raise click.Abort()

    if ctx.invoked_subcommand is None and command:
       __tiny_command(command)
       _upy.close()



@cli.command()
@click.argument("remote")
@click.argument("local", required=False)
def get(remote, local):
    """
    Download a file from the connected device to the local machine.
    :param remote: The remote file to download.
    :param local: The local file or directory to save the downloaded content. If not provided, print to stdout.
    """
    if not remote.startswith(_device_root_fs):
        remote = posixpath.join(_device_root_fs, remote)
    
    try:
        _upy.fs_get(remote, local)
    except UpyBoardError:
        remote = remote.replace(_device_root_fs, "", 1)
        print(f"The {ANSIEC.FG.BRIGHT_RED}{remote}{ANSIEC.OP.RESET} does not exist or is not a file.")

@cli.command()
@click.argument("remote")
def mkdir(remote):
    """
    Create a directory on the connected device.
    :param remote: The remote directory to create.
    """
    path_ = remote
    if not path_.startswith(_device_root_fs):
        path_ = _device_root_fs + remote
        
    if _upy.fs_mkdir(path_):
        print(f"{ANSIEC.FG.BRIGHT_BLUE}{remote}{ANSIEC.OP.RESET} is {ANSIEC.FG.BRIGHT_GREEN}created.{ANSIEC.OP.RESET}")
    else:
        print(f"{ANSIEC.FG.BRIGHT_BLUE}{remote}{ANSIEC.OP.RESET} is {ANSIEC.FG.BRIGHT_RED}already exists.{ANSIEC.OP.RESET}")

@cli.command()
@click.argument("remote")
def rm(remote):
    """
    Remove a file or directory from the connected device.
    :param remote: The remote file or directory to remove.
    """
    if not remote.startswith(_device_root_fs):
        remote = _device_root_fs + remote
        
    try:
        if _upy.fs_is_dir(remote):
            _upy.fs_rmdir(remote)
        else:
            _upy.fs_rm(remote)
    except UpyBoardError:
        remote = remote.replace(_device_root_fs, "", 1)
        print(f"The {ANSIEC.FG.BRIGHT_RED}{remote}{ANSIEC.OP.RESET} does not exist.")

def __get_icon_for_file(name, is_dir):
    """
    Return icon based on directory status or file extension.
    :param name: The name to get icon for.
    :param is_dir: True if the item is a directory.
    :return: Icon string for the file/directory.
    """
    if is_dir:
        return "📁"

    ext_icons = {
        ".py":   "🐍",
        ".mpy":  "📦",
        ".txt":  "📜",
        ".csv":  "📊",
        ".json": "🗄️",
    }
    _, ext = os.path.splitext(name.lower())
    return ext_icons.get(ext, "📄")

@cli.command()
@click.argument("path", default="/")
def ls(path):
    """
    List the files and directories in the specified path on the connected device,
    sorted and including file sizes and icons.
    :param path: The path to list. Defaults to the root directory ("/").
    """
    if not path.startswith(_device_root_fs):
        path = _device_root_fs + path

    try:
        items = _upy.fs_ls_detailed(path)
        
        if not items:
            print(f"{ANSIEC.FG.BRIGHT_RED}{path[1:]}{ANSIEC.OP.RESET} does not exist or is empty.")
            return
        
        display_items = []
        for name, size, is_dir in items:
            icon = __get_icon_for_file(name, is_dir)
            display_items.append((is_dir, name, size, icon))

        if display_items:
            size_width = max(len(str(item[2])) for item in display_items)

            for is_dir, f_name, size, icon in display_items:
                name_str = ANSIEC.FG.BRIGHT_BLUE + f_name + ANSIEC.OP.RESET if is_dir else f_name
                size_str = "" if is_dir else str(size)
                print(f"{size_str.rjust(size_width)}  {icon}  {name_str}")

    except UpyBoardError:
        print(f"{ANSIEC.FG.BRIGHT_RED}{path[1:]}{ANSIEC.OP.RESET} does not exist.")

@cli.command()
@click.argument("local", type=click.Path(exists=True))
@click.argument("remote", required=False)
def put(local, remote):
    """
    Upload a local file or directory to the connected device.
    :param local: The local file or directory to upload.
    :param remote: The remote path on the device. If not provided, the local file name will be used.
    """
    if remote is None:
        remote = os.path.basename(os.path.abspath(local))
    else:
        if not remote.startswith(_device_root_fs):
            remote = posixpath.join(_device_root_fs, remote)
        
        try:
            if _upy.fs_is_dir(remote):
                remote = remote + "/" + os.path.basename(os.path.abspath(local))
        except UpyBoardError:
            pass
        
    if os.path.isdir(local):
        _upy.fs_putdir(local, remote)
    else:
        _upy.fs_put(local, remote)

def __run_error_process(out, local_file):
    """
    Process the error output from the device and print it in a readable format.
    """
    print(f"{ANSIEC.OP.left(len(_error_header))}{ANSIEC.FG.BLACK}{'-'*40}Traceback{'-'*40}{ANSIEC.OP.RESET}")
    for l in out[1:-2]:
        if "<stdin>" in l:
            full_path = os.path.abspath(os.path.join(os.getcwd(), local_file))
            l = l.replace("<stdin>", full_path, 1)
        print(l.strip())
        
    try:
        err_line_raw = out[-2].strip()
        
        if "<stdin>" in err_line_raw:
            full_path = os.path.abspath(os.path.join(os.getcwd(), local_file))
            err_line = err_line_raw.replace("<stdin>", full_path, 1)
        else:
            match = re.search(r'File "([^"]+)"', err_line_raw)
            if match:
                device_src_path = os.path.join(_device_path, "src")
                full_path =  os.path.join(device_src_path, match.group(1))
                escaped_filename = re.sub(r"([\\\\])", r"\\\1", full_path)
                err_line = re.sub(r'File "([^"]+)"', rf'File "{escaped_filename}"', err_line_raw)
                
        print(f" {err_line}")
        
        err_content = out[-1].strip()

        match = re.search(r"line (\d+)", err_line)
        if match:
            line = int(match.group(1))
            try:
                with open(full_path, "r") as f:
                    lines = f.readlines()
                    print(f"  {lines[line - 1].rstrip()}")
            except:
                pass    

    except IndexError:
       err_content = out[-1].strip()
    

    print(f"{ANSIEC.FG.BRIGHT_MAGENTA}{err_content}{ANSIEC.OP.RESET}")
    
@cli.command()
@click.argument("local_file")
@click.option(
    "--not-waiting",
    "-n",
    is_flag=True,
    help="Don't wait for output.",
)
@click.option(
    "--input-echo",
    "-i",
    is_flag=True,
    help="Turn on echo for input",
)
def run(local_file, not_waiting, input_echo):
    """
    Run the local file on the connected device.
    :param local_file: The local file to run on the device.
    :param no_waiting: If True, do not wait for input/output stream.
    :param input_echo: If True, turn on echo for input.
    """
    try:
        _upy.run(local_file, not not_waiting, input_echo)
    except IOError:
        click.echo(f"File not found: {ANSIEC.FG.BRIGHT_RED + local_file + ANSIEC.OP.RESET}", err=True)
    except UpyBoardError as ex:
        __run_error_process(str(ex).strip().split('\n'), local_file)

@cli.command()
def repl():
    """
    Enter the REPL (Read-Eval-Print Loop) mode.
    """
    print(f"{ANSIEC.FG.MAGENTA}Entering REPL mode. Press Ctrl + C to exit{ANSIEC.OP.RESET}.")

    _upy.repl()


is_stop_formatting_process = None

def _formatting_process():
    while not is_stop_formatting_process:
        print(f"{ANSIEC.FG.BRIGHT_BLUE}.{ANSIEC.OP.RESET}", end="", flush=True)
        time.sleep(0.1)

@cli.command()
def format():
    """
    Format the file system of the connected device.
    """
    global is_stop_formatting_process

    print(f"Formatting the file system of {ANSIEC.FG.BRIGHT_YELLOW}{_device}{ANSIEC.OP.RESET}")
    
    is_stop_formatting_process = False
    th = threading.Thread(target=_formatting_process, daemon=True)
    th.start()
    ret = _upy.fs_format()
    is_stop_formatting_process = True
    th.join()
        
    if ret:
        print(f"{ANSIEC.OP.left()}{ANSIEC.OP.CLEAR_LINE}The file system has been {ANSIEC.FG.BRIGHT_BLUE}formatted{ANSIEC.OP.RESET}")
    else:
        print(f"{ANSIEC.OP.left()}The {ANSIEC.FG.BRIGHT_RED}{_device}{ANSIEC.OP.RESET} is not supported.")
    return ret

@cli.command()
def df():
    """
    Show the file system information of the connected device.
    """
    ret = _upy.fs_df()
    if ret:
        out_str = f"""Total: {ret[0]//1024:5} KByte ({ret[0]:5})
Used: {ret[1]//1024:6} KByte ({ret[1]:7})
Free: {ret[2]//1024:6} KByte ({ret[2]:6})
Usage: {round(ret[3],2):5} %""" 
        
        print(out_str)

@cli.command()
def shell():
    """
    Enter an interactive shell for device control.
    """
    import shlex

    COMMANDS = "clear, ls, cd, get, put, rm, mkdir, df, repl, pwd, help(?)"
    HELP = f"""Type 'exit' or press CTRL+C to exit.  
Available: {ANSIEC.FG.BRIGHT_BLUE}{COMMANDS}{ANSIEC.OP.RESET}"""
    
            
    current_path = '/' # _device_root_fs
    
    def print_prompt():
        print(f"\n📟 {_device}:{current_path} >", end=" ", flush=True)

    def run_cmd(cmdline):
        nonlocal current_path

        args = shlex.split(cmdline)
        if not args:
            return
        cmd = args[0]

        try:        
            if cmd == "ls":
                if len(args) > 1:
                    print("Usage: ls")
                    return

                click.Context(ls).invoke(ls, path=current_path)

            elif cmd == "cd":
                if len(args) != 2:
                    print("Usage: cd <dir>")
                    return
                
                new_path = posixpath.normpath(posixpath.join(current_path, args[1]))
                try:
                    _upy.fs_is_dir(new_path)
                    current_path = new_path
                except:
                    dirs = ANSIEC.FG.BRIGHT_RED + " ".join(args[1:]) + ANSIEC.OP.RESET
                    print(f"The {dirs} directory does not exist.")

            elif cmd == "get":
                if len(args) < 2 or len(args) > 3:
                    print("Usage: get <remote> [local]")
                    return
                remote = posixpath.join(current_path, args[1])
                local = args[2] if len(args) >= 3 else None
                click.Context(get).invoke(get, remote=remote, local=local)

            elif cmd == "put":
                if len(args) < 2 or len(args) > 3:
                    print("Usage: put <local> [remote]")
                    return
                
                local = args[1]
                remote = args[2] if len(args) >= 3 else None
                if remote is None:
                    remote = os.path.basename(local)
                remote = posixpath.join(current_path, remote)
                click.Context(put).invoke(put, local=local, remote=remote)

            elif cmd == "rm":
                if len(args) != 2:
                    print("Usage: rm <remote>")
                    return
                remote = posixpath.join(current_path, args[1])
                click.Context(rm).invoke(rm, remote=remote)

            elif cmd == "mkdir":
                if len(args) != 2:
                    print("Usage: mkdir <remote>")
                    return
                remote = posixpath.join(current_path, args[1])
                click.Context(mkdir).invoke(mkdir, remote=remote)

            elif cmd == "df":
                if len(args) > 1:
                    print("Usage: df")
                    return

                click.Context(df).invoke(df)

            elif cmd == "-c":
                if len(args) < 2:
                    print("Usage: -c <scripts>")
                    return
                
                scripts = cmdline[3:]
                __tiny_command(scripts)

            elif cmd == "repl":
                print("Press CTRL+C to exit.")
                if len(args) > 1:
                    print("Usage: repl")
                    return
                
                _upy.repl()

            elif cmd == "pwd":
                if len(args) > 1:
                    print("Usage: pwd")
                    return

                print(current_path)

            elif cmd == "clear":
                if len(args) > 1:
                    print("Usage: clear")
                    return
                print(ANSIEC.OP.CLEAR + ANSIEC.OP.RESET)

            elif cmd == "help" or cmd == "?":
                if len(args) > 1:
                    print("Usage: help or ?")
                    return
                
                print(HELP)
                
            else:
                raise Exception(f"Unknown command: {cmd}")
        except UpyBoardError:
            raise Exception(f"Unknown command: {cmdline}")
        
    print(f"Connected to {_device} on {ANSIEC.FG.BRIGHT_GREEN}{_core}{ANSIEC.OP.RESET}")
    print(HELP)
                
    try:
        while True:
            print_prompt()
            line = sys.stdin.buffer.readline().decode(errors='replace').rstrip()
            if line == 'exit':
                break
            try:
                run_cmd(line)
            except Exception as e:
                print(f"{e}")
                continue
            
    except (EOFError, KeyboardInterrupt):
        print()
    finally:
        print("Exiting shell.")
    
def __force_remove_readonly(func, path, exc_info):
    """
    Force remove a read-only file or directory.
    """
    try:
        os.chmod(path, stat.S_IWRITE)
        func(path)
    except Exception as e:
        print(f"Deletion failed: {path}, error: {e}")

@cli.command()
def reset():
    """
    Reset the connected device.
    """
    _upy.reset()
    
@cli.command()
@click.argument("device", required=False)
def env(device=None): 
    """
    Create a .vscode folder with the upyboard environment.
    """
    global _device, _device_path
        
    vscode_dir = ".vscode" 

    upyboard_env_file = os.path.join(vscode_dir, ".upyboard")
    task_file = os.path.join(vscode_dir, "tasks.json") 
    settings_file = os.path.join(vscode_dir, "settings.json") 
    launch_file = os.path.join(vscode_dir, "launch.json")
    
    old_upyboard_contents = None
    
    task_file_contents = """{
    "version": "2.0.0",
    "tasks": [
        {
            "label": "Run micropython with upyboard",
            "type": "shell",
            "command": "upy",
            "args": [
                "${file}"
            ],
            "problemMatcher": [],
            "group": {
                "kind": "build",
                "isDefault": true
            }
        }
    ]
}
"""

    settings_file_contents = """{
    "files.exclude": {
      "**/.vscode": true,
    },
    "python.languageServer": "Pylance",
    "python.analysis.diagnosticSeverityOverrides": {
        "reportMissingModuleSource": "none",
    },
    "python.analysis.extraPaths": [
        "./.vscode"
    ]
}
"""

    launch_file_contents = """{
    "version": "0.2.0",
    "configurations": [
      {
        "name": "Python: Current file debug",
        "type": "debugpy",
        "request": "launch",
        "program": "${file}",
        "console": "integratedTerminal"
      }
    ]
  }
"""
    if device:  
        leaf_values = [
            device_name
            for subdict in SUPPORT_CORE_DEVICE_TYPES.values()
            for device_name in subdict.values()
        ]
        if device in leaf_values:
            _device = device
            _device_path = os.path.join(os.path.dirname(upyboard.__file__), os.path.join("device", _device))
        else:
            print(f"The {ANSIEC.FG.BRIGHT_RED}{device}{ANSIEC.OP.RESET} is an unsupported device.")
            raise click.Abort()
            
    if os.path.exists(vscode_dir):
        upyboard_path = os.path.join(vscode_dir, ".upyboard")
        if os.path.isfile(upyboard_path):
            with open(upyboard_path, "r", encoding="utf-8") as f:
                old_upyboard_contents = f.read()
    
        print(f"There is already a set environment. Do you want to set it again? (y or n): ", end='', flush=True)
        while True:
            char = getch().lower()
            if char == b'n':
                print("\nCanceling the operation")
                return
            elif char == b'y':
                break
        
        shutil.rmtree(vscode_dir, onerror=__force_remove_readonly)

    core_typehints = os.path.join(_core_path, "typehints")
    shutil.copytree(core_typehints, vscode_dir) 
    
    if _device_path:
        vscode_device_dir = os.path.join(vscode_dir, _device)  
        device_typehints = os.path.join(_device_path, "typehints")  
        shutil.copytree(device_typehints, vscode_device_dir) 

    with open(upyboard_env_file, "w", encoding="utf-8") as f:
        if old_upyboard_contents is not None:
            f.write(old_upyboard_contents)
        else:
            f.write(f"SERIAL_PORT={_sport.upper()}\n")

    with open(task_file, "w", encoding="utf-8") as f:
        f.write(task_file_contents)  

    with open(settings_file, "w", encoding="utf-8") as f:
        f.write(settings_file_contents)

    with open(launch_file, "w", encoding="utf-8") as f:
        f.write(launch_file_contents)

    if old_upyboard_contents is None:
        print(f"Serial port {ANSIEC.FG.BRIGHT_GREEN}{_sport}{ANSIEC.OP.RESET} is registered.") 
    else:
        print(f"The registered serial port is {ANSIEC.FG.BRIGHT_GREEN}{old_upyboard_contents.split('=')[1][:-1]}{ANSIEC.OP.RESET}")

import mpy_cross

@cli.command()
@click.argument("local", type=click.Path(exists=True))
@click.argument("remote", required=False)
def upload(local, remote):
    """
    It compiles all source code from the local path into .mpy files and uploads them to the remote path.
    :param local: The local source code directory.
    :param remote: The remote path on the device. If not provided, root directory will be used.
    """
    def _mpy_output_path(base, filepath):
        relative_path = os.path.relpath(filepath, base)
        output_dir = os.path.join(os.path.dirname(base), "mpy", os.path.dirname(relative_path))
        os.makedirs(output_dir, exist_ok=True)
        
        filename = os.path.splitext(os.path.basename(filepath))[0] + ".mpy"
        return os.path.join(output_dir, filename)

    def _conv_py_to_mpy(local, base):
        args = ['_filepath_', '-o', '_outpath_', '-msmall-int-bits=31']
        if _core == "EFR32MG":
            if _version < 1.19:
                args.append('-mno-unicode')
        elif _core == "ESP32":
            args.append('-march=xtensa')
        elif _core == "RP2350":
            args.append('-march=armv7emsp')
        else:
            raise ValueError(f"The {ANSIEC.FG.BRIGHT_RED}{_core}{ANSIEC.OP.RESET} is not supported")
        
        if os.path.isfile(local):
            args[0] = local
            args[2] = os.path.splitext(local)[0] + ".mpy"    
            mpy_cross.run(*args)
        else:      
            for filename in os.listdir(local):
                filepath = os.path.join(local, filename)

                if os.path.isdir(filepath):
                    _conv_py_to_mpy(filepath, base)
                    continue

                if not filepath.endswith(".py"):
                    continue
    
                args[0] = filepath
                args[2] = _mpy_output_path(base, filepath)
                mpy_cross.run(*args)


    remote = _device_root_fs + (remote or "")
        
    _conv_py_to_mpy(local, base=local)
   
    if os.path.isfile(local):
        local = os.path.splitext(local)[0] + ".mpy"
        click.Context(put).invoke(put, local=local, remote=remote)
        os.remove(local)
    elif os.path.isdir(local):            
        shutil.rmtree(os.path.join(local, "__pycache__"), ignore_errors=True)

        local_mpy_dir = os.path.join(os.path.dirname(local), "mpy")
        if not os.path.exists(local_mpy_dir):
            return

        _upy.fs_mkdir(remote)

        for item in os.listdir(local_mpy_dir):
            local_item = os.path.join(local_mpy_dir, item)
            remote_item = os.path.join(remote, item).replace("\\", "/")
            click.Context(put).invoke(put, local=local_item, remote=remote_item)

        shutil.rmtree(local_mpy_dir, onerror=__force_remove_readonly)
    else:
        print("The local path is not a file or directory.")
    
@cli.command()
@click.argument("device", required=False)
def init(device=None):
    """
    Initialize the file system of the device.
    """
    global _device, _device_path
    
    if device:  
        leaf_values = [
            device_name
            for subdict in SUPPORT_CORE_DEVICE_TYPES.values()
            for device_name in subdict.values()
        ]
        if device in leaf_values:
            _device = device
            _device_path = os.path.join(os.path.dirname(upyboard.__file__), os.path.join("device", _device))
        else:
            print(f"The {ANSIEC.FG.BRIGHT_RED}{device}{ANSIEC.OP.RESET} is an unsupported device.")
            raise click.Abort()
        
    if _version < 1.12:
        print("Unkown micropython version " + ANSIEC.FG.BRIGHT_RED + f"{_version}" + ANSIEC.OP.RESET)
        raise click.Abort()
    
    if not click.Context(format).invoke(format):
        print("Unable to format the file system. Please check the " + ANSIEC.FG.BRIGHT_RED + f"{_device}" + ANSIEC.OP.RESET)
        return 
        
    _upy.fs_mkdir(_device_root_fs + "lib/")

    upload_format_str = (
        "Uploading the " 
        + ANSIEC.FG.BRIGHT_YELLOW + "{0}" + ANSIEC.OP.RESET 
        + " library on the " 
        + ANSIEC.FG.BRIGHT_YELLOW + "{1}" + ANSIEC.OP.RESET
    )
    
    core_src= os.path.join(_core_path, "src")
    if os.path.exists(core_src):
        print(upload_format_str.format("Core", _device))
        click.Context(upload).invoke(upload, local=core_src, remote="lib/")
    
    if _device_path:
        device_src = os.path.join(_device_path, "src")
        if os.path.exists(device_src):
            print(upload_format_str.format("Device", _device))
            click.Context(upload).invoke(upload, local=device_src, remote="lib/" + _device)

    print("The job is done!")


def __is_bluetooth_port(port_info):
    """
    Check if the given port_info is a Bluetooth port.
    :param port_info: The port information to check.
    :return: True if the port is a Bluetooth port, False otherwise.
    """
    bt_keywords = ['bluetooth', 'bth', 'devb', 'rfcomm', 'Blue', 'BT']
    description = port_info.description.lower()
    device = port_info.device.lower()
    return any(keyword in description or keyword in device for keyword in bt_keywords)

@cli.command()
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    default=False,
    help="Enable raw REPL mode for scanning",
)
def scan(raw:bool):
    """
    Display the list of connected boards.
    :param raw: If True, display detailed information about the connected device. otherwise, display the version and device name.
    """

    color_tbl = (ANSIEC.FG.BRIGHT_YELLOW, ANSIEC.FG.BRIGHT_GREEN, ANSIEC.FG.BRIGHT_BLUE)
    color_pos = 0    
    
    for port in list_ports.comports():
        if __is_bluetooth_port(port):
            continue
        descript = get_micropython_board_info(port.device, raw)
            
        if descript:
            color_pos = (color_pos + 1) % len(color_tbl)
            if not raw:
                version, date, core, device = descript
                                                 
                print(color_tbl[color_pos] + f"{port.device:>6}" + ANSIEC.OP.RESET + f"\t{version:>4} {date:>11}" + color_tbl[color_pos] + f"  {device}" + ANSIEC.OP.RESET)
            else:
                print(color_tbl[color_pos] + f"{port.device:>6}" + ANSIEC.OP.RESET + f"\t{descript}")


def __is_valid_serial_port(port_name:str):
    """
    Check if the port_name is valid on the specified or current platform.
    :param port_name: The serial port string to validate.
    :return: True if valid, False otherwise.
    """
    platform = sys.platform

    if platform.startswith("win"):
        return re.fullmatch(r"COM[1-9][0-9]*", port_name, re.IGNORECASE) is not None
    elif platform.startswith("linux"):
        return re.fullmatch(r"/dev/tty(USB|S|ACM)[0-9]+", port_name) is not None
    elif platform == "darwin":
        return re.fullmatch(r"/dev/tty\..+", port_name) is not None
    else:
        return False

@cli.command()
@click.argument("port", required=False)
def sport(port:str=None):
    """
    Set or display the serial port for the connected device.
    If no port is specified, it will read the port from the .vscode/.upyboard file.
    :param port: The serial port to set. If None, read from the .vscode/.upyboard file.
    """
    if port is None:
        if os.path.exists(os.path.join(".vscode", ".upyboard")):
            with open(os.path.join(".vscode", ".upyboard"), "r") as f:
                content = f.read()
                match = re.search(r"SERIAL_PORT=(.*)", content)
                if match:
                    port = match.group(1).strip()
                    print(f"Current serial port: {ANSIEC.FG.BRIGHT_GREEN}{port}{ANSIEC.OP.RESET}")
                else:
                    print("No serial port found.")
        else:
            print("No serial port is configured.")
    else:
        if not __is_valid_serial_port(port):
            print(f"Invalid serial port: {ANSIEC.FG.BRIGHT_RED}{port}{ANSIEC.OP.RESET}")
            return
        if not get_micropython_board_info(port):
            print(f"Device is not connected to {ANSIEC.FG.BRIGHT_RED}{port}{ANSIEC.OP.RESET}")
            return
        
        if os.path.exists(os.path.join(".vscode", ".upyboard")):
            with open(os.path.join(".vscode", ".upyboard"), "w") as f:
                f.write(f"SERIAL_PORT={port.upper()}\n")
            print(f"Serial port set to: {ANSIEC.FG.BRIGHT_GREEN}{port.upper()}{ANSIEC.OP.RESET}")
        else:
            print("Requires configuration.")

#--------------------------------------------------------------

UPDATE_INTERVAL = 60 * 60 * 24  
UPDATE_TIMESTAMP_FILE = pathlib.Path.home() / ".upyboard_update_check"


def __should_check_for_updates() -> bool:
    """
    Check if the update check should be performed based on the last check time.
    :return: True if the update check should be performed, False otherwise.
    """
    
    if UPDATE_TIMESTAMP_FILE.exists():
        last_check = UPDATE_TIMESTAMP_FILE.stat().st_mtime
        if time.time() - last_check < UPDATE_INTERVAL:
            return False
    return True

def check_for_updates(current_version):
    """
    Check PyPI for a newer version of upyboard and prompt the user to upgrade.
    """
    if not __should_check_for_updates():
        return

    try:
        with urllib.request.urlopen("https://pypi.org/pypi/upyboard/json") as resp:
            data = json.loads(resp.read().decode("utf-8"))
        latest_version = data["info"]["version"]

        if tuple(map(int, latest_version.split('.'))) > tuple(map(int, current_version.split('.'))):
            print(f"A newer version ({latest_version}) is available. Update now? (y/n): ", end='', flush=True)
            if getch().decode().lower() == 'y':
                subprocess.run([
                    sys.executable, "-m", "pip", "install",
                    "--upgrade", "upyboard",
                    "--upgrade-strategy", "eager"
                ])
                sys.exit(0)
    except Exception as e:
        pass
    
    try:
        UPDATE_TIMESTAMP_FILE.touch()
    except Exception:
        pass


def main():    
    check_for_updates(__version__)
    if len(sys.argv) == 1:
        print(f"Use {ANSIEC.FG.BRIGHT_BLUE}upy --help{ANSIEC.OP.RESET} to see available commands.")
        raise SystemExit()
        
    elif not any(item in sys.argv for item in ('get', 'put', 'rm', 'run', 'upload')) and sys.argv[-1].split('.')[-1] == 'py':
        index = next((i for i, arg in enumerate(sys.argv[1:], 1) if arg in ['-i', '--input-echo', '-n', '--not-waiting']), None)
        if index is not None:
            sys.argv.insert(index, 'run')
        else:
            sys.argv.insert(-1, 'run')

    load_env_from_upyboard()
    exit_code = cli()
    sys.exit(exit_code)
        	
if __name__ == '__main__':
    main()



import utime
import urandom
import math
import machine

class DistanceScanner:
    """
    A class to control a motor motor and an Distance sensor for distance scanning.
    It sweeps the Motor across a specified range and takes distance readings at each angle.
    The best angle with the minimum distance is returned.
    """
    
    def __init__(self, motor:object, sensor:object, step:int=2, settle_ms:int=20, samples:int=5):
        """
        :param motor: motorMotor object
        :param sensor: Ultrasensor object 
        :param step: int, angle step for motor motor
        :param settle_ms: int, time to wait for motor to settle in ms
        :param samples: int, number of samples to take for each angle
        """
        self.motor = motor
        self.sensor = sensor
        self.step = step
        self.settle_ms = settle_ms
        self.samples = samples

    def __median(self, lst:list) -> float:
        """
        calculate median value of a list of numbers.
        :param seq: list[float]
        :return: float
        """    
        lst = sorted(lst)
        n = len(lst)
        mid = n // 2
        return lst[mid] if n & 1 else 0.5 * (lst[mid-1] + lst[mid])

    def sweep(self, start:int=0, end:int=180) -> dict: 
        """
        Sweep the motor motor from start to end angle and return the best angle with minimum distance.
        :param start: int, starting angle
        :param end: int, ending angle
        :return: dict, best angle and distance
        """
        best = {'angle': None, 'dist': 1e9}
        direction = 1 if end >= start else -1

        for angle in range(start, end + direction, self.step * direction):
            self.motor.angle(angle)
            utime.sleep_ms(self.settle_ms)

            reads = []
            for _ in range(self.samples):
                d = self.sensor.read()
                if d is not None:
                    reads.append(d)
                utime.sleep_us(60)

            if reads:
                dist = self.__median(reads)
                if dist < best['dist']:
                    best = {'angle': angle, 'dist': dist}

        return best


class WS2812Matrix_Effect:
    def __init__(self, ws):
        """
        A class to apply various effects on a WS2812 LED matrix.
        It provides methods to create effects like sparkle, meteor rain, plasma, fireworks, campfire, and wave RGB.
        
        :param ws: WS2812Matrix object, the LED matrix to apply effects on
        """
        self.ws             = ws
        self.__timer        = None
        self.__state        = {}
        self.__effect_id    = 0
        self.__busy         = False

    def __install(self, period_s:float, handler):
        """
        Install a periodic callback to handle the effect.
        :param period_s: float, period in seconds for the effect
        :param handler: callable, the function to call at each period
        """
        self.stop()
        self.__effect_id += 1
        eid = self.__effect_id
        period_ms = int(period_s * 1000)

        def __cb(t):
            if eid != self.__effect_id or self.__busy:
                return
            self.__busy = True
            try:
                handler()
            finally:
                self.__busy = False

        tm = machine.Timer()
        tm.init(period=period_ms, mode=machine.Timer.PERIODIC, callback=__cb)
        self.__timer = tm

    def stop(self):
        """
        Stop the current effect and reset the effect ID.
        """
        self.__effect_id += 1
        if self.__timer:
            self.__timer.deinit()
            self.__timer = None

    def __wheel(self, pos:int):
        """
        Convert a position value to an RGB color using the color wheel algorithm.
        :param pos: int, position value (0-255)
        :return: tuple, RGB color
        """
        pos &= 255
        if pos < 85:
            return (255 - pos * 3, pos * 3, 0)
        if pos < 170:
            pos -= 85
            return (0, 255 - pos * 3, pos * 3)
        pos -= 170
        return (pos * 3, 0, 255 - pos * 3)

    def __heat_color(self, t:int):
        """
        Convert a temperature value to an RGB color.
        
        :param t: int, temperature value (0-255)
        :return: tuple, RGB color
        """
        if t <= 85:
            return (t * 3, 0, 0)
        if t <= 170:
            t -= 85
            return (255, t * 3, 0)
        t -= 170
        return (255, 255, t * 3)

    def sparkle(self, *, base=(0,0,0), sparkle_color=(255,255,255), density=0.1, decay=0.9, speed=0.03):
        """
        Create a sparkle effect on the LED matrix.
        
        :param base: tuple, RGB color for the base color of the matrix
        :param sparkle_color: tuple, RGB color for the sparkle
        :param density: float, density of the sparkles (0-1)
        :param decay: float, decay factor for the sparkle brightness (0-1)
        :param speed: float, speed of the effect in seconds
        """
        self.ws.fill(base)
        self.__state['spark'] = {'decay': decay, 'dens': density, 'sc': sparkle_color}
        self.__install(speed, self.__sparkle_step)

    def __sparkle_step(self):
        """
        Step function for the sparkle effect.
        It decays the brightness of existing sparkles and adds new sparkles based on the density.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        N = w * h
        s = self.__state['spark']
        for i in range(N):
            x, y = i % w, i // w
            r, g, b = ws[x, y]
            ws[x, y] = (int(r * s['decay']), int(g * s['decay']), int(b * s['decay']))
        if urandom.getrandbits(16) < int(65535 * s['dens']):
            idx = urandom.getrandbits(16) % N
            x, y = idx % w, idx // w
            ws[x, y] = s['sc']
        ws.update()

    def meteor_rain(self, *, colors=((255,0,0),(0,0,255)), count=3, decay=0.8, speed=0.04):
        """
        Create a meteor rain effect on the LED matrix.
        
        :param colors: tuple, list of RGB colors for the meteors
        :param count: int, number of meteors to create
        :param decay: float, decay factor for the meteor brightness (0-1)
        :param speed: float, speed of the effect in seconds
        """
        w, h = self.ws.display_w, self.ws.display_h
        N = w * h
        mets = [{
            'pos': urandom.getrandbits(16) % N,
            'spd': 1 + urandom.getrandbits(2),
            'col': colors[urandom.getrandbits(8) % len(colors)]
        } for _ in range(count)]
        self.__state['meteor'] = {'ms': mets, 'decay': decay}
        self.__install(speed, self.__meteor_step)

    def __meteor_step(self):
        """
        Step function for the meteor rain effect.
        It decays the brightness of existing pixels and updates the position and color of meteors.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        N = w * h
        s = self.__state['meteor']
        for i in range(N):
            x, y = i % w, i // w
            r, g, b = ws[x, y]
            ws[x, y] = (int(r * s['decay']), int(g * s['decay']), int(b * s['decay']))
        for m in s['ms']:
            x, y = m['pos'] % w, m['pos'] // w
            ws[x, y] = m['col']
            m['pos'] = (m['pos'] + m['spd']) % N
        ws.update()

    def plasma(self, *, hue_shift=2, speed=0.05):
        """
        Create a plasma effect on the LED matrix.
        
        :param hue_shift: int, the shift in hue for the plasma effect
        :param speed: float, speed of the effect in seconds
        """
        self.__state['plasma'] = {'t': 0, 'shift': hue_shift}
        self.__install(speed, self.__plasma_step)

    def __plasma_step(self):
        """
        Step function for the plasma effect.
        It calculates the color for each pixel based on a sine wave function and updates the matrix.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        st = self.__state['plasma']
        t = st['t']
        for y in range(h):
            for x in range(w):
                hval = (math.sin(x * 0.5 + t) + math.sin(y * 0.5 + t)) * 180 + t
                ws[x, y] = self.__wheel(int(hval) & 255)
        st['t'] += st['shift']
        ws.update()

    def fireworks(self, *, sparks=24, fade=0.9, speed=0.03, colors=((255,128,0),(255,255,255),(0,255,255))):
        """
        Create a fireworks effect on the LED matrix.
        
        :param sparks: int, number of sparks in the fireworks
        :param fade: float, fade factor for the sparks (0-1)
        :param speed: float, speed of the effect in seconds
        :param colors: tuple, list of RGB colors for the sparks
        """
        self.__state['fire'] = {'parts': [], 'fade': fade, 'colors': colors, 'cool': 0, 'sparks': sparks}
        self.__fire_spawn()
        self.__install(speed, self.__fire_step)

    def __fire_spawn(self):
        """
        Spawn new sparks for the fireworks effect.
        It clears existing sparks and generates new ones at a random position on the matrix.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        N = w * h
        s = self.__state['fire']
        s['parts'].clear()
        center = urandom.getrandbits(16) % N
        for _ in range(s['sparks']):
            vel = (urandom.getrandbits(3) % 5) + 1
            dir_ = 1 if urandom.getrandbits(1) else -1
            s['parts'].append({
                'pos': center,
                'vel': vel * dir_,
                'col': s['colors'][urandom.getrandbits(8) % len(s['colors'])]
            })

    def __fire_step(self):
        """
        Step function for the fireworks effect.
        It fades existing sparks, updates their positions, and spawns new sparks if necessary.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        N = w * h
        s = self.__state['fire']
        # fade
        for i in range(N):
            x, y = i % w, i // w
            r, g, b = ws[x, y]
            ws[x, y] = (int(r * s['fade']), int(g * s['fade']), int(b * s['fade']))
        alive = False
        for p in s['parts']:
            p['pos'] = (p['pos'] + p['vel']) % N
            x, y = p['pos'] % w, p['pos'] // w
            ws[x, y] = p['col']
            alive = True
        s['cool'] += 1
        if not alive or s['cool'] > 25:
            self.__fire_spawn()
            s['cool'] = 0
        ws.update()

    def campfire(self, *, cooling=55, sparking=120, speed=0.03):
        """
        Create a campfire effect on the LED matrix.
        
        :param cooling: int, cooling factor for the heat (0-255)
        :param sparking: int, sparking factor for the heat (0-255)
        :param speed: float, speed of the effect in seconds
        """
        w, h = self.ws.display_w, self.ws.display_h
        N = w * h
        self.__state['camp'] = {'heat': [0]*N, 'cool': cooling, 'spark': sparking}
        self.__install(speed, self.__camp_step)

    def __camp_step(self):
        """
        Step function for the campfire effect.
        It cools down the heat, drifts it up, and adds sparks randomly.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        s = self.__state['camp']
        heat = s['heat']
        N = w * h
        # cool down
        for i in range(N):
            cool = urandom.getrandbits(8) % ((s['cool'] * 10 // N) + 2)
            heat[i] = max(0, heat[i] - cool)
        # drift up
        for i in range(N-1, 1, -1):
            heat[i] = (heat[i-1] + heat[i-2] + heat[i-2]) // 3
        # spark
        if urandom.getrandbits(8) < s['spark']:
            idx = urandom.getrandbits(8) % min(3, N)
            heat[idx] = min(255, heat[idx] + urandom.getrandbits(8)%95 + 160)
        # map to color
        for i in range(N):
            x, y = i % w, i // w
            ws[x, y] = self.__heat_color(heat[i])
        ws.update()

    def wave_rgb(self, *, speed=0.1):
        """
        Create a wave RGB effect on the LED matrix.
        
        :param speed: float, speed of the effect in seconds
        """
        self.__state['wave'] = {'step': 0}
        self.__install(speed, self.__wave_step)

    def __wave_step(self):
        """
        Step function for the wave RGB effect.
        It calculates the color for each pixel based on a sine wave function and updates the matrix.
        """
        ws = self.ws
        w, h = ws.display_w, ws.display_h
        s = self.__state['wave']
        step = s['step']
        N = w * h
        for i in range(N):
            base = step + i * (360 / N)
            r = int((math.sin(math.radians(base)) + 1)/2 * 255)
            g = int((math.sin(math.radians(base+120)) + 1)/2 * 255)
            b = int((math.sin(math.radians(base+240)) + 1)/2 * 255)
            x, y = i % w, i // w
            ws[x, y] = (r, g, b)
        s['step'] = (step + 5) % 360
        ws.update()


"""
Module: 'sys' on micropython-v1.25.0-rp2-RPI_PICO2_W
"""

# MCU: {'build': '', 'ver': '1.25.0', 'version': '1.25.0', 'port': 'rp2', 'board': 'RPI_PICO2_W', 'mpy': 'v6.3', 'family': 'micropython', 'cpu': 'RP2350', 'arch': 'armv7emsp'}
# Stubber: v1.24.0
from __future__ import annotations
from _typeshed import Incomplete

platform: str = "rp2"
version_info: tuple = ()
path: list = []
version: str = "3.4.0; MicroPython v1.25.0 on 2025-04-15"
ps1: str = ">>> "
ps2: str = "... "
byteorder: str = "little"
modules: dict = {}
argv: list = []
implementation: tuple = ()
maxsize: int = 2147483647

def print_exception(*args, **kwargs) -> Incomplete: ...
def exit(*args, **kwargs) -> Incomplete: ...

stderr: Incomplete  ## <class 'TextIOWrapper'> = <io.TextIOWrapper 2>
stdout: Incomplete  ## <class 'TextIOWrapper'> = <io.TextIOWrapper 1>
stdin: Incomplete  ## <class 'TextIOWrapper'> = <io.TextIOWrapper 0>



"""
Module: 'aioble.__init__' on micropython-v1.25.0-rp2-RPI_PICO2_W
"""

# MCU: {'build': '', 'ver': '1.25.0', 'version': '1.25.0', 'port': 'rp2', 'board': 'RPI_PICO2_W', 'mpy': 'v6.3', 'family': 'micropython', 'cpu': 'RP2350', 'arch': 'armv7emsp'}
# Stubber: v1.24.0
from __future__ import annotations
from typing import Final, Generator
from _typeshed import Incomplete

ADDR_PUBLIC: Final[int] = 0
ADDR_RANDOM: Final[int] = 1

def log_warn(*args, **kwargs) -> Incomplete: ...
def log_info(*args, **kwargs) -> Incomplete: ...
def log_error(*args, **kwargs) -> Incomplete: ...
def register_services(*args, **kwargs) -> Incomplete: ...
def stop(*args, **kwargs) -> Incomplete: ...
def config(*args, **kwargs) -> Incomplete: ...
def const(*args, **kwargs) -> Incomplete: ...

class Service:
    def _tuple(self, *args, **kwargs) -> Incomplete: ...
    def __init__(self, *argv, **kwargs) -> None: ...

class scan:
    def cancel(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def __init__(self, *argv, **kwargs) -> None: ...

def advertise(*args, **kwargs) -> Generator:  ## = <generator>
    ...

class Characteristic:
    def _remote_write(self, *args, **kwargs) -> Incomplete: ...
    def _remote_read(self, *args, **kwargs) -> Incomplete: ...
    def notify(self, *args, **kwargs) -> Incomplete: ...
    def on_read(self, *args, **kwargs) -> Incomplete: ...
    def _tuple(self, *args, **kwargs) -> Incomplete: ...
    def write(self, *args, **kwargs) -> Incomplete: ...
    def read(self, *args, **kwargs) -> Incomplete: ...
    def _register(self, *args, **kwargs) -> Incomplete: ...
    def _indicate_done(self, *args, **kwargs) -> Incomplete: ...
    def _init_capture(self, *args, **kwargs) -> Incomplete: ...
    def written(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def indicate(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def _run_capture_task(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def __init__(self, *argv, **kwargs) -> None: ...

class GattError(Exception): ...

class BufferedCharacteristic:
    def on_read(self, *args, **kwargs) -> Incomplete: ...
    def _remote_read(self, *args, **kwargs) -> Incomplete: ...
    def _remote_write(self, *args, **kwargs) -> Incomplete: ...
    def notify(self, *args, **kwargs) -> Incomplete: ...
    def _tuple(self, *args, **kwargs) -> Incomplete: ...
    def _init_capture(self, *args, **kwargs) -> Incomplete: ...
    def read(self, *args, **kwargs) -> Incomplete: ...
    def write(self, *args, **kwargs) -> Incomplete: ...
    def _indicate_done(self, *args, **kwargs) -> Incomplete: ...
    def written(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def _register(self, *args, **kwargs) -> Incomplete: ...
    def indicate(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def _run_capture_task(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def __init__(self, *argv, **kwargs) -> None: ...

class DeviceDisconnectedError(Exception): ...

class Descriptor:
    def _tuple(self, *args, **kwargs) -> Incomplete: ...
    def _remote_read(self, *args, **kwargs) -> Incomplete: ...
    def _remote_write(self, *args, **kwargs) -> Incomplete: ...
    def on_read(self, *args, **kwargs) -> Incomplete: ...
    def _register(self, *args, **kwargs) -> Incomplete: ...
    def read(self, *args, **kwargs) -> Incomplete: ...
    def write(self, *args, **kwargs) -> Incomplete: ...
    def _init_capture(self, *args, **kwargs) -> Incomplete: ...
    def written(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def _run_capture_task(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def __init__(self, *argv, **kwargs) -> None: ...

class Device:
    def addr_hex(self, *args, **kwargs) -> Incomplete: ...
    def connect(*args, **kwargs) -> Generator:  ## = <generator>
        ...

    def __init__(self, *argv, **kwargs) -> None: ...





def clamp(val:int|float, lo:int|float, hi:int|float) -> int|float:
    """
    Clamps a value between a lower and upper bound.\r
\1 val: The value to be clamped.
    :param lo: The lower bound.
    :param hi: The upper bound.
    :return: The clamped value.
    """
    
def map(x:int|float, min_i:int|float, max_i:int|float, min_o:int|float, max_o:int|float) -> int|float:
    """
    Maps a value from one range to another.\r
\1 x: The value to be mapped.
    :param min_i: The minimum value of the input range.
    :param min_o: The minimum value of the output range.
    :return: The mapped value.
    """

def xrange(start: float, stop: float | None = None, step: float | None = None) -> iter[float]:
    """
    A generator function to create a range of floating point numbers.
    This is a replacement for the built-in range function for floating point numbers.   
    :param start: Starting value of the range.
    :param stop: Ending value of the range.
    :param step: Step size for the range.
    :return: A range object that generates floating point numbers.
    """

def rand(size:int=4) -> int:
    """
    Generates a random number of the specified size in bytes.\r
\1 size: The size of the random number in bytes. Default is 4 bytes.
    :return: A random number of the specified size.
    """
        
def intervalChecker(interval_ms:int) -> callable:
    """
    Creates a function that checks if the specified interval has passed since the last call.\r
\1 interval_ms: The interval in milliseconds.
    :return: A function that checks if the interval has passed.
    """


class ANSIEC:
    """
    ANSI Escape Codes for terminal text formatting.
    This class provides methods for setting foreground and background colors, as well as text attributes.
    It uses ANSI escape codes to format text in the terminal.
    """
    
    class FG:
        """
        ANSI escape codes for foreground colors.
        This class provides methods for setting foreground colors using ANSI escape codes.
        It includes standard colors, bright colors, and RGB color support.
        """
        
        BLACK = "\u001b[30m"
        RED = "\u001b[31m"
        GREEN = "\u001b[32m"
        YELLOW = "\u001b[33m"
        BLUE = "\u001b[34m"
        MAGENTA = "\u001b[35m"
        CYAN = "\u001b[36m"
        WHITE = "\u001b[37m"
        BRIGHT_BLACK= "\u001b[30;1m"
        BRIGHT_RED = "\u001b[31;1m"
        BRIGHT_GREEN = "\u001b[32;1m"
        BRIGHT_YELLOW = "\u001b[33;1m"
        BRIGHT_BLUE = "\u001b[34;1m"
        BRIGHT_MAGENTA = "\u001b[35;1m"
        BRIGHT_CYAN = "\u001b[36;1m"
        BRIGHT_WHITE = "\u001b[37;1m"
                
        @classmethod
        def rgb(cls, r:int, g:int, b:int) -> str:
            """
            Returns an ANSI escape code for RGB foreground color.
            :param r: Red component (0-255).
            :param g: Green component (0-255).
            :param b: Blue component (0-255).
            :return: An ANSI escape code for RGB foreground color.
            """

    class BG:
        """
        ANSI escape codes for background colors.
        This class provides methods for setting background colors using ANSI escape codes.
        It includes standard colors, bright colors, and RGB color support.
        """
        
        BLACK = "\u001b[40m"
        RED = "\u001b[41m"
        GREEN = "\u001b[42m"
        YELLOW = "\u001b[43m"
        BLUE = "\u001b[44m"
        MAGENTA = "\u001b[45m"
        CYAN = "\u001b[46m"
        WHITE = "\u001b[47m"
        BRIGHT_BLACK= "\u001b[40;1m"
        BRIGHT_RED = "\u001b[41;1m"
        BRIGHT_GREEN = "\u001b[42;1m"
        BRIGHT_YELLOW = "\u001b[43;1m"
        BRIGHT_BLUE = "\u001b[44;1m"
        BRIGHT_MAGENTA = "\u001b[45;1m"
        BRIGHT_CYAN = "\u001b[46;1m"
        BRIGHT_WHITE = "\u001b[47;1m"
                
        @classmethod
        def rgb(cls, r:int, g:int, b:int) -> str:
            """
            Returns an ANSI escape code for RGB background color.\r
\1 r: Red component (0-255).
            :param g: Green component (0-255).
            :param b: Blue component (0-255).
            :return: An ANSI escape code for RGB background color.
            """

    class OP:
        """
        A class for managing ANSI escape codes for font attributes and cursor positioning.
        This class provides methods to control font styles and cursor movement using ANSI escape codes.
        It supports actions such as resetting attributes, applying bold, underline, and reverse effects, clearing the screen or lines, and moving the cursor.
        """
        
        RESET = "\u001b[0m"
        BOLD = "\u001b[1m"
        UNDER_LINE = "\u001b[4m"
        REVERSE = "\u001b[7m"
        CLEAR = "\u001b[2J"
        CLEAR_LINE = "\u001b[2K"
        TOP = "\u001b[0;0H"

        @classmethod
        def up(cls, n:int) -> str:
            """
            Cursor up\r
\1 n: Number of lines to move up.
            :return: An ANSI escape code to move the cursor up.
            """

        @classmethod
        def down(cls, n:int) -> str:
            """
            Cursor down\r
\1 n: Number of lines to move down.
            :return: An ANSI escape code to move the cursor down.
            """
            
        @classmethod
        def right(cls, n:int) -> str:
            """
            Cursor right\r
\1 n: Number of columns to move right.
            :return: An ANSI escape code to move the cursor right.
            """

        @classmethod
        def left(cls, n:int) -> str:
            """
            Cursor left\r
\1 n: Number of columns to move left.
            :return: An ANSI escape code to move the cursor left.
            """
           
        @classmethod
        def next_line(cls, n:int) -> str:
            """
            Cursor down to next line\r
\1 n: Number of lines to move down.
            :return: An ANSI escape code to move the cursor down.
            """

        @classmethod
        def prev_line(cls, n:int) -> str:
            """
            Cursor up to previous line\r
\1 n: Number of lines to move up.
            :return: An ANSI escape code to move the cursor up.
            """
        
        @classmethod
        def to(cls, row:int, colum:int) -> str:
            """
            Move cursor to specified row and column.\r
\1 row: Row number (1-based).
            :param colum: Column number (1-based).
            :return: An ANSI escape code to move the cursor.
            """

 
class SlipEncoder:
    """
    SLIP Encoder class.
    This class is used to encode a byte array into a SLIP frame.
    The SLIP frame is a byte array that starts and ends with the END byte (0xC0).
    The ESC byte (0xDB) is used to escape the END and ESC bytes in the payload.
    The ESC byte is followed by ESC_END (0xDC) for END and ESC_ESC (0xDD) for ESC.
    """

    @staticmethod
    def encode(payload: bytes) -> bytes:
        """"
        Encode a byte array into a SLIP frame.
        The SLIP frame is a byte array that starts and ends with the END byte (0xC0).
        The ESC byte (0xDB) is used to escape the END and ESC bytes in the payload.
        The ESC byte is followed by ESC_END (0xDC) for END and ESC_ESC (0xDD) for ESC.\r
\1 payload: The byte array to encode.
        :return: The encoded SLIP frame as a byte array.
        """


class SlipDecoder:
    """
    SLIP Decoder class.
    This class is used to decode SLIP frames. It handles the decoding of the SLIP protocol,
    including escaping and unescaping bytes.
    """

    def __init__(self) -> None:
        """
        Initialize the SLIP decoder.
        This sets up the internal buffer and flags for processing SLIP frames.
        """

    def reset(self) -> None:
        """
        Reset the decoder state.
        This clears the buffer and resets the escape and in-frame flags.
        """

    def feed(self, chunk: bytes) -> list[bytes]:
        """
        Feed a chunk of bytes to the decoder.
        This processes the chunk and returns a list of complete SLIP frames.
        If the chunk contains incomplete frames, they will be buffered for the next call.\r
\1 chunk: The chunk of bytes to process.
        :return: A list of complete SLIP frames as byte arrays.
        """


class RingBuffer:
    """
    A simple ring buffer implementation for byte data.
    This class provides methods to put data into the buffer, get data from it,
    check available data, and peek for a specific pattern.
    """
    def __init__(self, size:int):
        """
        Initialize the ring buffer with a specified size.\r
\1 size: The size of the ring buffer.
        """

    def put(self, data:bytes):
        """
        Put data into the ring buffer.
        If the buffer is full, it will overwrite the oldest data.\r
\1 data: The data to put into the buffer (must be bytes).
        """

    def avail(self) -> int:
        """
        Check the number of available bytes in the ring buffer.
        
        :return: The number of bytes available in the buffer.
        """

    def get(self, n:int=1) -> bytes:
        """
        Get `n` bytes from the ring buffer.
        If `n` is greater than the available bytes, it will return only the available bytes.
        If `n` is None, it will return all available bytes.\r
\1 n: The number of bytes to get (default is 1).
        :return: The bytes read from the buffer.
        """
    
    def peek(self, n:int = 1) -> bytes:
        """
        Peek `n` bytes from the ring buffer without removing them.
        If `n` is greater than the available bytes, it will return only the available bytes.
        If `n` is None, it will return all available bytes.\r
\1 n: The number of bytes to peek (default is 1).
        :return: The bytes peeked from the buffer.
        """
        
    def get_until(self, pattern:bytes, max_size:int|None=None) -> bytes|None:
        """
        Peek for a specific pattern in the ring buffer.
        Searches for the first occurrence of `pattern` in the buffer.
        If `max_size` is specified, it will return at most `max_size` bytes.
        If the pattern is not found, it returns None.\r
\1 pattern: The byte sequence to search for.
        :param max_size: The maximum size of data to return (default is None, no limit).
        :return: The bytes up to the pattern if found, otherwise None.
        """


# Copyright (c) 2019, Digi International, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, ContextManager, Iterator, Optional

from . import modem_status

ADDR_BROADCAST: bytes = ...
ADDR_COORDINATOR: bytes = ...
ENDPOINT_DIGI_DATA: int = ...
CLUSTER_DIGI_SERIAL_DATA: int = ...
PROFILE_DIGI_XBEE: int = ...

PIN_WAKE: int = ...
RTC_WAKE: int = ...

def atcmd(cmd: str, value: Optional[Any] = None) -> Optional[Any]:
    """
    Sets or queries an AT command on the XBee module.\r
\1 cmd: Two-character string that represents the command.
    :param value: Command value.

        * If the ``value`` parameter is not set, the function executes the AT
          command and, depending on the command, returns the result as either
          a string, bytes object, an integer, or ``None``. Some commands
          simply return a value; other AT commands, such as special commands
          and execution commands, change the behavior of the XBee device. For
          example, **FR** resets the device.
        * If the ``value`` parameter is set, you can specify a value in a
          string, bytearray, or integer format. The function passes the value
          to set the AT command.

    :return: The value of the queried AT command. The format of the value
        depends on the requested AT command.
    """
    ...

def discover() -> Iterator[dict]:
    """
    Performs a network discovery, which is equivalent to issuing the **ND**
    command.

    Method returns immediately, but querying the resulting iterator will block
    execution until a response is available or the discovery times out (as
    determined by **N?**).

    **Note**: This method only applies to XBee3 RF modules. XBee Cellular
        modules do not support discover functionality.

    :return: A dictionary for each discovered node that contains the following
        entries:

        * ``sender_nwk`` - 16-bit network address.
        * ``sender_eui64`` - 8-byte bytes object with EUI-64 address.
        * ``parent_nwk`` - Set to **0xFFFE** on the coordinator and routers,
          otherwise the network address of the end device's parent.
        * ``node_id`` - The device's **NI** value (a string of up to 20
          characters, also referred to as Node Identification).
        * ``node_type`` - Value of **0**, **1** or **2** for coordinator,
          router or end device.
        * ``device_type`` - The device's 32-bit **DD** value (also referred to
          as Digi Device Type).
        * ``rssi`` - RSSI of the node discovery request packet received by the
          sending node.
    """
    ...

def receive() -> dict:
    """
    Returns an entry from the receive queue. The format and fields are
    equivalent to receiving a **0x91** Explicit Rx API frame.

    If the device is operating in MicroPython REPL (**AP** is set to **4**) and
    the receive queue is full, it silently rejects any additional incoming
    packets; the sending node will receive a transmission status of **0x24**
    (Address not found) in this case.

    It's recommended to call the ``receive()`` function in a loop so no data is
    lost. On devices where there is a high volume of network traffic, there
    could be data lost if the messages are not pulled from the queue fast
    enough.

    **Note**: This method only applies to XBee3 RF modules. XBee Cellular
        modules do not support receive functionality.

    :return: A received data entry as a dictionary containing the following
        fields:

        * ``sender_nwk``: The 16-bit network address of the sending node
        * ``sender_eui64``: The 64-bit address (as a bytearray) of the sending
          node
        * ``source_ep`` - The source endpoint as an integer.
        * ``dest_ep`` - The destination endpoint as an integer.
        * ``cluster`` - The cluster id as an integer.
        * ``profile`` - The profile id as an integer.
        * ``broadcast`` - Either ``True`` or ``False`` depending on whether the
          frame was broadcast or unicast.
        * ``payload`` - A bytes object of the payload (intentional selection of
          bytes object over string since the payload can contain binary data).
    """
    ...

def transmit(dest: Any, payload: Any, *, source_ep: int = xbee.ENDPOINT_DIGI_DATA,
             dest_ep: int = xbee.ENDPOINT_DIGI_DATA, cluster: int = xbee.CLUSTER_DIGI_SERIAL_DATA,
             profile: int = xbee.PROFILE_DIGI_XBEE, bcast_radius: int = 0, tx_options: int = 0) -> None:
    """
    Transmits a packet to a specified destination address. This function either
    succeeds and returns ``None``, or raises an exception. Here is a partial
    list of the exceptions to expect:

    * ``TypeError`` - Invalid type for either ``dest`` or ``payload``
    * ``ValueError`` - Payload is too long. Maximum length depends on whether
      you are making a unicast or broadcast transmission with or without
      encryption. Note that application-level encryption is not available in
      current builds.
    * ``OSError(ENOTCONN)`` - Device is not joined to a network (**AI** returns
      a non-zero value)
    * ``OSError(EAGAIN)`` - Temporary issue preventing sending, for example,
      insufficient buffers, packet already queued for target
    * ``OSError(EIO)`` - General error message for **unable to send**

    **Note**: This method only applies to XBee3 RF modules. XBee Cellular
        modules do not support transmit functionality.\r
\1 dest: The destination address of the message, and accepts any of the
        following:

        * An integer for 16-bit addressing.
        * An 8-byte bytes object for 64-bit addressing.
        * The constant ``xbee.ADDR_BROADCAST`` to indicate a broadcast
          destination.
        * The constant ``xbee.ADDR_COORDINATOR`` to indicate the coordinator.

        There are multiple ways to create the 8-byte bytes object for 64-bit
        addressing:

        * As a bytestring: ``b'\\x00\\x13\\xa2\\x00\x41\\x74\\x07\\xa6'``.
        * Using the ``bytes()`` constructor with a list of decimal values:
          ``bytes([0, 19, 162, 0, 65, 116, 7, 166])``.
        * Using the ``bytes()`` constructor with a tuple of hex values:
          ``bytes((0x00, 0x13, 0xa2, 0x00, 0x41, 0x74, 0x07, 0xa6))``.
    :param payload: A string (for example, 'Hello World!') or bytes object
        (useful for sending binary data).
    :param source_ep: Optional 8-bit Source Endpoint for the transmission,
        defaulting to ``xbee.ENDPOINT_DIGI_DATA``.
    :param dest_ep: Optional 8-bit Destination Endpoint for the transmission,
        defaulting to ``xbee.ENDPOINT_DIGI_DATA``.
    :param cluster: Optional 16-bit Cluster ID for the transmission,
        defaulting to ``xbee.CLUSTER_DIGI_SERIAL_DATA``.
    :param profile: Optional 16-bit Cluster ID for the transmission,
        defaulting to ``xbee.PROFILE_DIGI_XBEE``.
    :param bcast_radius: Optional 8-bit value to set the maximum number of
        hops a broadcast transmission can traverse. Default is 0.
    :param tx_options: Optional 8-bit bitfield that configures advanced
        transmission options. Please see the protocol specific user manual
        for TX Options usage.

    **Note**: All of the optional parameters are new in the following firmware versions:
        * XBee3 Zigbee: version 1007
        * XBee3 802.15.4: version 2004
        * XBee3 DigiMesh 2.4: version 3003
    """
    ...

def idle_radio(idle: bool, /) -> None:
  """
  Enters or exits the idle radio state. `idle_now(True)` puts the device
  into this state, and `idle_now(False)` returns it to normal operation.
  While the radio is idle, the module consumes less current but cannot
  receive RF transmissions.

  **Note**: This function is only available on XBee 3 RF modules with
  version x00B or newer. XBee Cellular modules do not support this
  functionality.
  """
  ...

def poll_now() -> None:
  """
  Polls for messages sent while the device was asleep or the radio idle.

  In order to receive data while the radio is idled, the MicroPython
  application must periodically call xbee.poll_now() to check for
  incoming data. Calling poll_now() will cause the radio to check for
  messages stored by its parent, and will retrieve a single message from
  the parent if any are available. xbee.poll_now() doesn't return the
  retrieved packet—the packet is instead passed to MicroPython via the
  normal method, either by the application calling xbee.receive() or
  through a callback if one is registered.

  **Note**: xbee.poll_now() is a non-blocking call, meaning it will
  return before the polling is complete. It can take about 10 ms after
  calling poll_now() before a message is available to xbee.receive().

  It is recommended to register a callback for received packets when
  using xbee.poll_now(). That way, once the packet is retrieved from the
  parent it will be immediately handled by the application.

  Since xbee.poll_now() only retrieves a single packet from the parent,
  it may need to be called more than once if more than one message may
  be received between polls. One way to do so is by calling
  xbee.poll_now() again every time a packet is received, or in the
  receive callback.

  **Note**: This function is only available on XBee3 Zigbee modules with
  version 100B or newer.
  """
  ...

class XBee:
    """
    Class used to create an object for the XBee device that is hosting
    MicroPython.
    """

    wake_lock: ContextManager = ...

    def __init__(self) -> None:
        """
        Class constructor. Instantiates a new ``XBee`` object.
        """
        ...

    def atcmd(self, cmd: str, value: Optional[Any] = None) -> Optional[Any]:
        """
        Sets or queries an AT command on the XBee3 RF Module.\r
\1 cmd: Two-character string that represents the command.
        :param value: Command value.

            * If the ``value`` parameter is not set, the function executes the
              AT command and, depending on the command, returns the result as
              either a string, bytes object, an integer, or ``None``. Some
              commands simply return a value; other AT commands, such as
              special commands and execution commands, change the behavior of
              the XBee device. For example, **FR** resets the device.
            * If the ``value`` parameter is set, you can specify a value in a
              string, bytearray, or integer format. The function passes the
              value to set the AT command.

        :return: The value of the queried AT command. The format of the value
            depends on the requested AT command.
        """
        ...

    def sleep_now(self, timeout_ms: int, pin_wake: bool=False) -> int:
        """
        Puts the XBee device in sleep mode. The device sleeps for the specified
        time period programmed with an optional early pin wake (DTR,
        commissioning button, or SPI_SSEL).

        Notice that sleep mode must be disabled by setting **SM** (Sleep Mode)
        to **0**, in order to control when the module sleeps.

        Throws an ``EALREADY`` OSError exception if **SM** is already
        configured for sleep (set to something other than **0**).\r
\1 timeout_ms: The number of milliseconds the device will sleep.
        :param pin_wake: If set to ``True``, the device only goes to sleep if
            **DIO8** is pulled high.

        :return: The number of milliseconds elapsed.
        """
        ...

    def wake_reason(self) -> int:
        """
        Returns the reason that made the radio to wake up. One of:

        * ``PIN_WAKE`` - If the **full** ``timeout_ms`` elapsed.
        * ``RTC_WAKE`` - When ``pin_wake`` is enabled and **DIO8** woke the
          device early.

        :return: The reason that made the radio to wake up.
        """
        ...



from . import Dout


class Relay:
    ON = Dout.HIGH
    OFF = Dout.LOW

    class _Channel:
        """
        A class representing a single relay channel.
        This class provides properties to get and set the state of the relay,
        turn the relay on or off, and set the delay in milliseconds.
        """
        def __init__(self, parent:Relay, idx:int):
            """
            Initialize the relay channel with a reference to the parent Relay instance and its index.

            :param parent: Reference to the parent Relay instance
            :param idx: Index of the relay channel (0-based)
            """

        @property
        def state(self) -> int:
            """
            Get the current state of the relay channel.
            This property returns the state of the relay channel, which can be either Relay.ON or Relay.OFF.
            
            :return: Current state of the relay channel (Dout.HIGH or Dout.LOW)
            """

        @state.setter
        def state(self, val:int):
            """
            Set the state of the relay channel.
            This property sets the state of the relay channel to either Relay.ON or Relay.OFF.

            :param val: State to set (Dout.HIGH or Dout.LOW)
            """

        def on(self):
            """
            Turn the relay channel on.
            This method sets the state of the relay channel to Relay.ON.
            """

        def off(self):
            """
            Turn the relay channel off.
            This method sets the state of the relay channel to Relay.OFF.
            """

        @property
        def delay_ms(self) -> int:
            """
            Get the minimum delay in milliseconds for the relay channel.
            This property returns the minimum delay in milliseconds that is applied after changing the state of the relay channel.
            
            :return: Minimum delay in milliseconds
            """

        @delay_ms.setter
        def delay_ms(self, ms:int):
            """
            Set the minimum delay in milliseconds for the relay channel.
            This property sets the minimum delay in milliseconds that is applied after changing the state of the relay channel.
            
            :param ms: Minimum delay in milliseconds
            """

    class _Group:
        """
        A class representing a group of relay channels.
        This class allows setting attributes for all channels in the group at once.
        """
        def __init__(self, channels:tuple[int, ...]):
            """
            Initialize the group of relay channels.
            
            :param channels: Tuple of Relay._Channel objects representing the relay channels in the group
            """

        def __setattr__(self, name, value):
            """
            Set an attribute for all channels in the group.
            If the attribute name starts with an underscore, it is set on the group itself.
            
            :param name: Attribute name
            :param value: Value to set for the attribute
            """

        def __getattr__(self, name):
            """
            Get an attribute from the first channel in the group.
            """

    def __init__(self, pins:tuple[int, ...], min_delay_ms:int=5):
        """
        Initialize the Relay with specified GPIO pins and minimum delay.

        :param pins: Tuple of GPIO pin numbers for the relay channels (e.g., (2, 3, 4))
        :param min_delay_ms: Minimum delay in milliseconds after changing the state of a relay channel (default 5 ms)
        """

    def __getitem__(self, idx:int) -> _Channel:
        """
        Get the relay channel at the specified index.

        :param idx: Index of the relay channel (0-based)
        :return: Relay._Channel object for the specified index
        """

    def __len__(self) -> int:
        """
        Get the number of relay channels.
        
        :return: Number of relay channels
        """


class ServoMotor:
    """
    A class to control servo motors using PWM.
    This class allows setting the angle of individual servo motors and provides
    methods to set the angle for all servos at once.
    """
    class _Channel:
        """
        A class representing a single servo channel.
        This class provides properties to get and set the angle, speed, and non-blocking behavior of the servo.
        """
        def __init__(self, parent:ServoMotor, idx:int):
            """
            Initialize the servo channel with a reference to the parent ServoMotor instance and its index.
            :param parent: Reference to the parent ServoMotor instance
            :param idx: Index of the servo channel (0-based)
            """

        @property
        def angle(self) -> float:
            """
            Get the current angle of the servo channel.
            This property returns the current angle in degrees (0 to 180).
            
            :return: Current angle in degrees
            """

        @angle.setter
        def angle(self, deg:float) -> None:
            """
            Set the target angle for the servo channel.
            This property clamps the angle to the range [0, 180] degrees and updates the target angle.
            
            :param deg: Target angle in degrees (0 to 180)
            """

        @property
        def speed_ms(self) -> int:
            """
            Get the speed in milliseconds for the servo channel.
            This property returns the speed in milliseconds for moving to the target angle.
            
            :return: Speed in milliseconds
            """

        @speed_ms.setter
        def speed_ms(self, ms:int):
            """
            Set the speed in milliseconds for the servo channel.
            This property sets the speed for moving to the target angle.
            
            :param ms: Speed in milliseconds (0 for immediate movement)
            """

        @property
        def nonblocking(self) -> bool:
            """
            Get the non-blocking flag for the servo channel.
            This property indicates whether the servo channel operates in non-blocking mode.
            
            :return: True if non-blocking mode is enabled, False otherwise
            """

        @nonblocking.setter
        def nonblocking(self, flag: bool) -> None:
            """
            Set the non-blocking flag for the servo channel.
            This property enables or disables non-blocking mode for the servo channel.
            
            :param flag: True to enable non-blocking mode, False to disable
            """

    class _Group:
        def __init__(self, channels:tuple[int, ...]):
            """
            Initialize the group of servo channels.
            
            :param channels: List of ServoMotor._Channel objects
            """

        def __setattr__(self, name:str, value:int):
            """
            Set an attribute for all channels in the group.
            If the attribute name starts with an underscore, it is set on the group itself.
            
            :param name: Attribute name
            :param value: Value to set for the attribute
            """

        def __getattr__(self, name):
            """
            Get an attribute from the first channel in the group.
            If the attribute does not exist, it raises an AttributeError.
            
            :param name: Attribute name
            :return: Value of the attribute from the first channel
            """

    def __init__(self, pins:tuple[int, ...], freq:int=50, default_min_us:int=500, default_max_us: int=2500, initial_angle: float=0.0):
        """
        Initialize the ServoMotor with specified GPIO pins and parameters.
        
        :param pins: Tuple of GPIO pin numbers for the servo motors (e.g., (2, 3, 4))
        :param freq: PWM frequency in Hz (default 50 Hz)
        :param default_min_us: Default minimum pulse width in microseconds (default 500 us)
        :param default_max_us: Default maximum pulse width in microseconds (default 2500 us)
        :param initial_angle: Initial angle for all servos in degrees (default 0.0)
        """
        self.__channels = [ServoMotor._Channel(self, i) for i in range(len(pins))]
        self.all = ServoMotor._Group(self.__channels)

    def __getitem__(self, idx: int) -> _Channel:
        """
        Get the servo channel at the specified index.
        
        :param idx: Index of the servo channel (0-based)
        :return: ServoMotor._Channel object for the specified index
        """
        
    def deinit(self) -> None:
        """
        Deinitialize the ServoMotor instance.
        This method stops the timer and disables all PWM channels.
        """


class PiezoBuzzer:
    """
    A class to control a piezo buzzer using PWM.
    It can play tones, melodies, and supports effects like staccato, vibrato, and tremolo.
    """

    def __init__(self, pin:int, tempo:int=120):
        """
        Initialize the PiezoBuzzer with the specified pin and tempo.
        
        :param pin: GPIO pin number for the buzzer (default 1).
        :param tempo: Tempo in beats per minute (default 120).
        """

    def tone(self, note_oct:str, length:int=4, effect:str=None):
        """
        Play a single tone with the specified note, length, and effect.
        
        :param note_oct: Musical note with octave (e.g., 'C4', 'A#5').
        :param length: Length of the note in beats (default 4).
        :param effect: Effect to apply to the note (e.g., 'staccato', 'vibrato', 'tremolo', 'gliss:C#5').
        """

    def play(self, melody, effect:str=None, background:bool=False):
        """
        Play a melody consisting of notes and lengths.
        
        :param melody: List of notes and lengths (e.g., ['C4', 4, 'D4', 2, 'E4', 1]).
        :param effect: Effect to apply to the melody (e.g., 'staccato', 'vibrato', 'tremolo').
        :param background: If True, play the melody in the background (default False).
        """

    def stop(self):
        """
        Stop playing the current melody and reset the buzzer state.
        """

    def set_tempo(self, bpm:int):
        """
        Set the tempo for the buzzer in beats per minute.
        
        :param bpm: Tempo in beats per minute
        """


class SR04:
    """
    This class drives an HC-SR04 ultrasonic sensor by emitting a 40 kHz pulse 
    and measuring its time-of-flight (using the speed of sound ≈343 m/s at 20 °C) 
    to compute distances from 2 cm to 400 cm, then applies a Kalman filter 
    to smooth out measurement noise.
    """    
    
    def __init__(self, trig:int, echo:int, *, temp_c:float=20.0, R:int=25, Q:int=4):
        """
        Initialize the ultrasonic sensor with the specified trigger and echo pins.
        
        :param trig: GPIO pin number for the trigger pin.
        :param echo: GPIO pin number for the echo pin.
        :param temp_c: Temperature in degrees Celsius (default is 20.0).
        :param R: Measurement noise covariance (default is 25).
        :param Q: Process noise covariance (default is 4).
        """
    
    def read(self, timeout_us:int=30_000, temp_c:float|None=None) -> float|None:
        """
        Read the distance from the ultrasonic sensor.
        
        :param timeout_us: Timeout in microseconds for the echo signal.
        :return: Distance in centimeters or None if timeout occurs.
        """


class WS2812Matrix:
    def __init__(self, pin_sm_pairs:list[tuple[int,int]], panel_w:int, panel_h:int, grid_w:int,  grid_h:int, *, zigzag:bool=True, origin:str='top_left', brightness:float=0.25):
        """
        WS2812 Matrix controller using PIO for multiple panels.
        
        :param pin_sm_pairs: list of (pin_number, state_machine_id) tuples
        :param panel_w: width of each panel in pixels
        :param panel_h: height of each panel in pixels
        :param grid_w: number of panels in the grid width
        :param grid_h: number of panels in the grid height
        :param zigzag: if True, odd rows are reversed (zigzag wiring)
        :param origin: one of 'top_left', 'top_right', 'bottom_left', 'bottom_right'
        :param brightness: brightness level from 0.0 to 1.0
        """

    def deinit(self):
        """
        Deinitialize the WS2812 matrix controller.
        This method clears the buffers, stops the state machines, and resets the pins.
        """

    @property
    def display_w(self) -> int:
        """
        Get the width of the display in pixels.
        
        :return: Width of the display in pixels
        """

    @property
    def display_h(self) -> int:
        """
        Get the height of the display in pixels.
        
        :return: Height of the display in pixels
        """

    @property
    def brightness(self) -> float:
        """
        Get the current brightness level of the matrix.
        The brightness level is a float value between 0.0 (off) and 1.0 (full brightness).
        
        :return: Brightness level (float)
        """

    @brightness.setter
    def brightness(self, value:float):
        """
        Set the brightness level of the matrix.
        The brightness level should be a float value between 0.0 (off) and 1.0 (full brightness).
        
        :param value: Brightness level (float)
        :raises ValueError: If the brightness value is not between 0.0 and 1.0.
        """

    def __setitem__(self, pos:tuple[int,int], color:tuple[int,int,int]):
        """
        Set the color of a pixel at the specified position.
        The color should be a tuple of three integers representing the RGB values (0-255).
        
        :param pos: Tuple of (x, y) coordinates of the pixel
        :param color: Tuple of (R, G, B) color values
        :raises IndexError: If the pixel coordinates are out of range.
        """

    def __getitem__(self, pos) -> tuple[int,int,int]:
        """
        Get the color of a pixel at the specified position.
        The color is returned as a tuple of three integers representing the RGB values (0-255).
        
        :param pos: Tuple of (x, y) coordinates of the pixel
        :return: Tuple of (R, G, B) color values
        :raises IndexError: If the pixel coordinates are out of range.
        """

    def fill(self, color:tuple[int,int,int]):
        """
        Fill the entire display with the specified color.
        The color should be a tuple of three integers representing the RGB values (0-255).
        
        :param color: Tuple of (R, G, B) color values
        :raises ValueError: If the color values are not in the range 0-255.
        """

    def clear(self):
        """
        Clear the display by filling it with black (0, 0, 0).
        This method sets all pixels to black and updates the display.
        """
        
    def update(self):
        """
        Update the display by sending the pixel data to the state machines.
        This method processes the pixel data in the buffers and sends it to the WS2812 panels.
        """

    def draw_line(self, x0:int, y0:int, x1:int, y1:int, color:tuple[int,int,int]):
        """
        Draw a line on the display from (x0, y0) to (x1, y1) with the specified color.
        
        :param x0: Starting x-coordinate of the line
        :param y0: Starting y-coordinate of the line
        :param x1: Ending x-coordinate of the line
        :param y1: Ending y-coordinate of the line
        :param color: Tuple of (R, G, B) color values
        """

    def draw_rect(self, x:int, y:int, w:int, h:int, color:tuple[int,int,int]):
        """
        Draw a rectangle on the display with the specified position, width, height, and color.
        
        :param x: x-coordinate of the top-left corner of the rectangle
        :param y: y-coordinate of the top-left corner of the rectangle
        :param w: Width of the rectangle
        :param h: Height of the rectangle
        :param color: Tuple of (R, G, B) color values
        """

    def draw_circle(self, cx:int, cy:int, r:int, color:tuple[int,int,int]):
        """
        Draw a circle on the display with the specified center, radius, and color.
        
        :param cx: x-coordinate of the center of the circle
        :param cy: y-coordinate of the center of the circle
        :param r: Radius of the circle
        :param color: Tuple of (R, G, B) color values
        """

    def blit(self, data, dst_x:int, dst_y:int, size_x:int, size_y:int, fg:tuple[int, int, int]=(255,255,255), bg:tuple[int, int, int]|None=None):
        """
        Blit a bitmap image onto the display at the specified position.
        The bitmap data can be a 2D list or a flat array of bytes.
        Each pixel is represented by a single byte (0 for black, 1 for white).
        
        :param data: Bitmap data as a 2D list or flat array of bytes
        :param dst_x: x-coordinate of the top-left corner where the bitmap will be drawn
        :param dst_y: y-coordinate of the top-left corner where the bitmap will be drawn
        :param size_x: Width of the bitmap in pixels
        :param size_y: Height of the bitmap in pixels
        :param fg: Foreground color as a tuple of (R, G, B) values (default is white)
        :param bg: Background color as a tuple of (R, G, B) values (default is None, which means no background)
        :raises ValueError: If the data is not in the expected format.
        """


class tLed:
    RED = (255, 0, 0)
    GREEN = (0, 255, 0)
    BLUE = (0, 0, 255)
    YELLOW = (255, 255, 0)
    CYAN = (0, 255, 255)
    MAGENTA = (255, 0, 255)
    WHITE = (255, 255, 255)

    def __init__(self, bright:float=1.0):
        """
        Basic WS2812 control class built into TiCLE.
        This class provides methods to turn on the LED with a specified color,
        turn it off, and define some common colors.
        """
                
    def on(self, color:tuple[int,int,int]=RED):
        """
        Turn on the LED with the specified color.

        :param color: Tuple of (R, G, B) color values (default is red).
        """

    def off(self):
        """
        Turn off the LED by filling it with black (0, 0, 0).
        """
        
    @property
    def brightness(self) -> float:
        """
        Get the current brightness level of the matrix.
        The brightness level is a float value between 0.0 (off) and 1.0 (full brightness).
        
        :return: Brightness level (float)
        """

    @brightness.setter
    def brightness(self, value:float):
        """
        Set the brightness level of the matrix.
        The brightness level should be a float value between 0.0 (off) and 1.0 (full brightness).
        
        :param value: Brightness level (float)
        :raises ValueError: If the brightness value is not between 0.0 and 1.0.
        """





"""
Module: 'ssl' on micropython-v1.25.0-rp2-RPI_PICO2_W
"""

# MCU: {'build': '', 'ver': '1.25.0', 'version': '1.25.0', 'port': 'rp2', 'board': 'RPI_PICO2_W', 'mpy': 'v6.3', 'family': 'micropython', 'cpu': 'RP2350', 'arch': 'armv7emsp'}
# Stubber: v1.24.0
from __future__ import annotations
from typing import Final
from _typeshed import Incomplete

PROTOCOL_TLS_SERVER: Final[int] = 1
PROTOCOL_DTLS_CLIENT: Final[int] = 2
PROTOCOL_DTLS_SERVER: Final[int] = 3
PROTOCOL_TLS_CLIENT: Final[int] = 0
MBEDTLS_VERSION: Final[str] = "Mbed TLS 3.6.2"
CERT_NONE: Final[int] = 0
CERT_OPTIONAL: Final[int] = 1
CERT_REQUIRED: Final[int] = 2

def wrap_socket(*args, **kwargs) -> Incomplete: ...

class SSLContext:
    def load_verify_locations(self, *args, **kwargs) -> Incomplete: ...
    def wrap_socket(self, *args, **kwargs) -> Incomplete: ...
    def load_cert_chain(self, *args, **kwargs) -> Incomplete: ...

    verify_mode: Incomplete  ## <class 'property'> = <property>
    def __init__(self, *argv, **kwargs) -> None: ...



"""
Module: 'network' on micropython-v1.25.0-rp2-RPI_PICO2_W
"""

# MCU: {'build': '', 'ver': '1.25.0', 'version': '1.25.0', 'port': 'rp2', 'board': 'RPI_PICO2_W', 'mpy': 'v6.3', 'family': 'micropython', 'cpu': 'RP2350', 'arch': 'armv7emsp'}
# Stubber: v1.24.0
from __future__ import annotations
from typing import Final
from _typeshed import Incomplete

STA_IF: Final[int] = 0
STAT_IDLE: Final[int] = 0
STAT_NO_AP_FOUND: Final[int] = -2
STAT_WRONG_PASSWORD: Final[int] = -3
STAT_GOT_IP: Final[int] = 3
AP_IF: Final[int] = 1
STAT_CONNECTING: Final[int] = 1
STAT_CONNECT_FAIL: Final[int] = -1

def hostname(*args, **kwargs) -> Incomplete: ...
def ipconfig(*args, **kwargs) -> Incomplete: ...
def route(*args, **kwargs) -> Incomplete: ...
def country(*args, **kwargs) -> Incomplete: ...

class WLAN:
    PM_POWERSAVE: Final[int] = 17
    SEC_OPEN: Final[int] = 0
    SEC_WPA2_WPA3: Final[int] = 20971524
    SEC_WPA3: Final[int] = 16777220
    SEC_WPA_WPA2: Final[int] = 4194310
    PM_PERFORMANCE: Final[int] = 10555714
    IF_AP: Final[int] = 1
    IF_STA: Final[int] = 0
    PM_NONE: Final[int] = 16
    def ipconfig(self, *args, **kwargs) -> Incomplete: ...
    def status(self, *args, **kwargs) -> Incomplete: ...
    def ioctl(self, *args, **kwargs) -> Incomplete: ...
    def send_ethernet(self, *args, **kwargs) -> Incomplete: ...
    def isconnected(self, *args, **kwargs) -> Incomplete: ...
    def scan(self, *args, **kwargs) -> Incomplete: ...
    def config(self, *args, **kwargs) -> Incomplete: ...
    def ifconfig(self, *args, **kwargs) -> Incomplete: ...
    def active(self, *args, **kwargs) -> Incomplete: ...
    def disconnect(self, *args, **kwargs) -> Incomplete: ...
    def connect(self, *args, **kwargs) -> Incomplete: ...
    def deinit(self, *args, **kwargs) -> Incomplete: ...
    def __init__(self, *argv, **kwargs) -> None: ...



import utime
import ustruct
import machine

from micropython import const


class Pir:
    NONE = 0
    DETECT = 1
    ENTER = 2
    LEAVE = 3
    
    def __init__(self):
        self.__pin = machine.Pin('P2', machine.Pin.IN)
        self.__t0 = 0
        
    def read(self):
        return self.__pin.value()    

    def state(self, timeout=1):
        old_state = self.read()
        utime.sleep_ms(timeout)
        curr_state = self.read()
        if old_state != curr_state:
            return self.ENTER if curr_state else self.LEAVE
        else:
            return curr_state
    
    def detect(self, timeout=1500):
        n = self.read()
        if n == self.DETECT and not self.__t0:
            self.__t0 = utime.ticks_ms()
        elif self.__t0:
            if utime.ticks_ms() - self.__t0 >= timeout:
                self.__t0 = 0
            n = not self.DETECT
        return n == self.DETECT

class IRThermometer:
    MLX90614_ADDR = const(0x5A)

    EEPROM_PWMCTRL = const(0x02)
    EEPROM_CONFIG_REGISTER1 = const(0x05)

    def __init__(self):
        self.__i2c = machine.I2C(1, freq=50000) #The maximum frequency of the MLX90614 is 100 KHz and the minimum is 10 KHz. 

    def read(self, object=True, eeprom=False):
        reg = const(0x07) if object else const(0x06)
        if eeprom:
            reg = 0x20 | reg
            
        data = self.__i2c.readfrom_mem(IRThermometer.MLX90614_ADDR, reg, 2)
        return ustruct.unpack('<H', data)[0]

    def ambient(self):
        data = self.read(False)
        return round(data * 0.02 - 273.15, 1)

    def object(self):
        data = self.read()
        return round(data * 0.02 - 273.15, 1)


class IMU:
    BNO055_ADDR = const(0x28)

    ACCELERATION = const(0x08)
    MAGNETIC = const(0x0E)
    GYROSCOPE = const(0x14)
    EULER = const(0x1A)
    QUATERNION = const(0x20)
    ACCEL_LINEAR = const(0x28)
    ACCEL_GRAVITY = const(0x2E)
    TEMPERATURE = const(0x34)
    
    def __init__(self):
        self.__i2c = machine.I2C(1)
        self.__scale = {self.ACCELERATION:1/100, self.MAGNETIC:1/16, self.GYROSCOPE:0.001090830782496456, self.EULER:1/16,  self.QUATERNION:1/(1<<14), self.ACCEL_LINEAR:1/100, self.ACCEL_GRAVITY:1/100}
        self.__call = {self.ACCELERATION:self.__read_other, self.MAGNETIC:self.__read_other, self.GYROSCOPE:self.__read_other, self.EULER:self.__read_other,  self.QUATERNION:self.__read_quaternion, self.ACCEL_LINEAR:self.__read_other, self.ACCEL_GRAVITY:self.__read_other, self.TEMPERATURE:self.__read_temperature}

        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0X3D, bytes([0x00])) #Mode Register, Enter configuration.
        utime.sleep_ms(20)
        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0x3F, bytes([0x20])) #Trigger Register, Reset
        utime.sleep_ms(650)
        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0X3E, bytes([0x00])) #Power Register, Set to normal power. cf) low power is 0x01
        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0X07, bytes([0x00])) #Page Register, Make sure we're in config mode and on page0(param, data), page1(conf)
        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0X3F, bytes([0x80])) #Trigger Register, External oscillator
        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0X3F, bytes([0x00])) #Trigger Register,
        utime.sleep_ms(10)
        self.__i2c.writeto_mem(IMU.BNO055_ADDR, 0X3D, bytes([0x0C])) #Mode Register, Enter normal operation (NDOF)
        utime.sleep_ms(200)

    def __read_temperature(self, addr):
        t = self.__i2c.readfrom_mem(IMU.BNO055_ADDR, addr, 1)[0]
        return t - 256 if t > 127 else t

    def __read_quaternion(self, addr):
        t = self.__i2c.readfrom_mem(IMU.BNO055_ADDR, addr, 8)  
        return tuple(v * self.__scale[self.QUATERNION] for v in ustruct.unpack('hhhh', t))

    def __read_other(self, addr):
        if addr not in self.__scale:
            raise ValueError(f"Address {addr} not in scale mapping")
        t = self.__i2c.readfrom_mem(IMU.BNO055_ADDR, addr, 6)
        return tuple(v * self.__scale[addr] for v in ustruct.unpack('hhh', t))

    def calibration(self):
        data = self.__i2c.readfrom_mem(IMU.BNO055_ADDR, 0x35, 1)[0] #Calibration Resiger, Read        
        return (data >> 6) & 0x03, (data >> 4) & 0x03, (data >> 2) & 0x03, data & 0x03  #Sys, Gyro, Accel, Mag

    def read(self, addr):
        return self.__call[addr](addr)


class _I2CtoUart:
    SC16IS750_ADDR = const(0x4D)
    
    # Register addresses (using datasheet table 21 as a reference)
    RHR = const(0x00)  # Receive Holding Register (read)
    THR = const(0x00)  # Transmit Holding Register (write)
    IER = const(0x01)  # Interrupt Enable Register
    FCR = const(0x02)  # FIFO Control Register
    IIR = const(0x02)  # Interrupt Identification Register
    LCR = const(0x03)  # Line Control Register
    MCR = const(0x04)  # Modem Control Register
    LSR = const(0x05)  # Line Status Register
    MSR = const(0x06)  # Modem Status Register
    SPR = const(0x07)  # Scratchpad Register
    TXLVL = const(0x08)  # Transmit FIFO Level
    RXLVL = const(0x09)  # Receive FIFO Level
    IODIR = const(0x0A)  # IO pin direction control
    IOSTATE = const(0x0B) # IO pin state control
    IOINTENA = const(0x0C) # IO interrupt enable
    IOCTRL = const(0x0E)  # IO control configuration
    EFCR = const(0x0F)  # Extra features control register

    DLL = const(0x00)  # Divisor Latch LSB (when LCR[7] = 1)
    DLH = const(0x01)  # Divisor Latch MSB (when LCR[7] = 1)

    # Line Status Register (LSR) bits
    LSR_DR = const(0x01)  # Data Ready
    LSR_OE = const(0x02)  # Overrun Error
    LSR_PE = const(0x04)  # Parity Error
    LSR_FE = const(0x08)  # Framing Error
    LSR_BI = const(0x10)  # Break Interrupt
    LSR_THRE = const(0x20)  # Transmitter Holding Register Empty
    LSR_TEMT = const(0x40)  # Transmitter Empty
    LSR_FIFOE = const(0x80) # FIFO Error

    # FIFO Control Register (FCR) bits
    FCR_FIFO_ENABLE = const(0x01)  # FIFO enable
    FCR_FIFO_RCVR_RESET = const(0x02) # Reset Receiver FIFO
    FCR_FIFO_XMIT_RESET = const(0x04) # Reset Transmitter FIFO
    FCR_FIFO_DMA_MODE = const(0x08)  # DMA mode
    FCR_FIFO_64 = const(0x20) # Enable 64 byte FIFO (SC16IS750/760)
    FCR_FIFO_RCVR_TRIGGER_MASK = const(0xC0)  # Receiver FIFO trigger level
    FCR_FIFO_RCVR_TRIGGER_8 = const(0x00)  # 8 bytes
    FCR_FIFO_RCVR_TRIGGER_16 = const(0x40) # 16 bytes
    FCR_FIFO_RCVR_TRIGGER_32 = const(0x80) # 32 bytes
    FCR_FIFO_RCVR_TRIGGER_60 = const(0xC0) # 60 bytes

    # Modem Control Register (MCR) bits
    MCR_DTR = const(0x01)  # DTR output
    MCR_RTS = const(0x02)  # RTS output
    MCR_LOOP = const(0x10) # Loop back mode
        
    def __init__(self, baudrate=9600, data_bits=8, stop_bits=1, parity='N'):
        self._i2c = machine.I2C(1)
        
        """
        Initializes the UART communication parameters.\r
\1 baudrate: The desired baud rate (default: self.baudrate).
        :param data_bits: The number of data bits (default: 8).
        :param stop_bits: The number of stop bits (default: 1).
        :param parity: The parity (default: 'N' - None, 'E' - Even, 'O' - Odd).
        """

        crystal_freq = 12_000_000   # XNode crystal frequency: 12MHz
        
        # Calculate divisor
        divisor = int(crystal_freq / (baudrate * 16))
        
        self.reset()
        
        # Enable access to divisor registers
        lcr = self.__read(self.LCR)
        self.__write(self.LCR, lcr | 0x80)

        # Write divisor
        self.__write(self.DLL, divisor & 0xFF)           # Division Register LSB
        self.__write(self.DLH, (divisor >> 8) & 0xFF)    

        # Set data bits, stop bits, and parity
        lcr = 0
        if data_bits == 8:
            lcr |= 0x03
        elif data_bits == 7:
            lcr |= 0x02
        elif data_bits == 6:
            lcr |= 0x01
        elif data_bits == 5:
            lcr |= 0x00
        else:
            raise ValueError("Invalid number of data bits")

        if stop_bits == 2:
            lcr |= 0x04

        if parity == 'E':
            lcr |= 0x18
        elif parity == 'O':
            lcr |= 0x08        

        # Disable divisor access and set communication parameters
        self.__write(self.LCR, lcr & 0x7F)

        # Configure FIFO
        self.__write(self.FCR, self.FCR_FIFO_ENABLE | \
                                  self.FCR_FIFO_RCVR_RESET | \
                                  self.FCR_FIFO_XMIT_RESET | \
                                  self.FCR_FIFO_RCVR_TRIGGER_16)

        # Enable RX interrupts
        self.__write(self.IER, 0x01)

    def __read(self, reg):
        return self._i2c.readfrom_mem(self.SC16IS750_ADDR, reg, 1)[0]

    def __write(self, reg, data):
        self._i2c.writeto_mem(self.SC16IS750_ADDR, reg, bytes([data]))

    def reset(self):
        """Resets the SC16IS750."""
        # Software reset via IOControl register
        self.__write(self.IOCTRL, 0x02)
        utime.sleep_ms(1)
    
    def available(self):
        """Returns the number of bytes available to read."""
        return self.__read(self.RXLVL)
            
    def read(self, size=1):
        """Reads data from the RX FIFO.\r
\1 size: The number of bytes to read (default: 1).

        :return: The bytes type data read from the RX FIFO.
        """
        
        if size == 1:
            if self.available() > 0:
                return bytes([self.__read(self.RHR)])
            else:
                return bytes()
        else:
            data = bytearray()
            for _ in range(size):
                if self.available() > 0:
                    data.append(self.__read(self.RHR))
                else:
                    break
            return bytes(data)
        
    def write(self, data):
        """Writes data to the TX FIFO.

        param data: The data to write (bytes or str).
        """
        if isinstance(data, str):
            data = data.encode('utf-8')

        for byte in data:
            while not (self._read(self.LSR) & self.LSR_THRE):
                pass  # Wait until TX FIFO is not full
            self.__write(self.THR, byte)

class Gps:
    """
    GPS module control class.
    """
    # FGPMMOPA6H(MediaTek MT3339) GPS Module
    # NMEA Checksum Calculator: https://nmeachecksum.eqth.net/
    
    UPDATE_1HZ = "$PMTK220,1000*1F"
    UPDATE_2HZ = "$PMTK220,200*2C"
    UPDATE_10HZ = "$PMTK220,100*2F"

    GPGGA = "$PMTK314,0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"
    GPVTG = "$PMTK314,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"
    GPRMC = "$PMTK314,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0*29"    
    
    BAUD_9600   = "$PMTK251,9600*17"
    BAUD_19200  = "$PMTK251,19200*22"
    BAUD_38400  = "$PMTK251,38400*27"   #Do Not!!!
    BAUD_115200 = "$PMTK251,115200*1F"  #Do Not!!!
        
    def __init__(self, first_update=True, gps_mode=GPGGA, baudrate=BAUD_9600):
        """
        Initializes the Gps object.

        Args:
            i2c: The I2C object from the machine module.
            i2c_address: The I2C address of the SC16IS750 (default: 0x4D).
            sc16is750_crystal_freq: The crystal oscillator frequency of the SC16IS750 (default: 11059200).
            baudrate: The baud rate for communication with the GPS module (default: 9600).
        """
        if baudrate == self.BAUD_9600:
            i2c_uart_bourate = 9600
        elif baudrate == self.BAUD_19200:
            i2c_uart_bourate = 19200
        elif i2c_uart_bourate == self.BAUD_38400:
            i2c_uart_bourate = 38400
        elif i2c_uart_bourate == self.BAUD_115200:
            i2c_uart_bourate = 115200
        else:
            raise ValueError("Invalid baudrate")
         
        self.__uart = _I2CtoUart(i2c_uart_bourate)
        self.setBaudrate(baudrate)
                
        self.setFastUpdate(first_update)
        self.setGpsMode(gps_mode)
        
        self.latest_nmea = ""
        self.latitude = 0.0
        self.longitude = 0.0
        self.altitude = 0.0
        self.fix_quality = 0
        self.satellites_in_use = 0
        self.timestamp = None
        self.speed_over_ground_knots = 0.0
        self.speed_over_ground_kmh = 0.0
        self.course_over_ground = 0.0
        self.fix_status = 'V' # Default to 'V' (void/invalid)
        
    def __sendCommand(self, command):
        """
        Sends a command to the GPS module.\r
\1 command: The command string to send.
        """
        self.__uart.write(command + '\r\n')
        utime.sleep_ms(50)

    def setBaudrate(self, baudrate):
        self.__sendCommand(baudrate)

    def setFastUpdate(self, first_update):
        self.__sendCommand(self.UPDATE_2HZ if first_update else self.UPDATE_1HZ) 
    
    def setGpsMode(self, gps_mode):
        self.gps_mode = gps_mode
        self.__sendCommand(gps_mode) 

    def update(self, timeout_ms = 2000):
        """
        Reads NMEA sentences from the GPS module and parses them.\r
\1 timeout_ms: Maximum time in milliseconds to wait for a complete NMEA sentence.
            
        :return: True if a valid NMEA sentence has been parsed, False otherwise
        """

        start_time = utime.ticks_ms()
        while (utime.ticks_ms() - start_time) < timeout_ms:
            if self.__uart.available():
                nmea_char = self.__uart.read().decode("ascii")
                self.latest_nmea += nmea_char
                if nmea_char == "\n":
                    if self._parse_nmea(self.latest_nmea):
                        self.latest_nmea = ""
                        return True
                    else:
                        self.latest_nmea = ""
                        return False
        return False

    def __parse_nmea_gpgga(self, nmea_sentence):
        if nmea_sentence.startswith("$GPGGA"):
            parts = nmea_sentence.split(",")
            if len(parts) >= 10 and all(parts[i] for i in [1, 2, 3, 4, 5, 6, 9]):  # Check for empty fields
                try:
                    # Time (UTC)
                    self.timestamp = parts[1]

                    # Latitude
                    lat_degrees = int(parts[2][:2])
                    lat_minutes = float(parts[2][2:])
                    self.latitude = lat_degrees + lat_minutes / 60.0
                    if parts[3] == "S":
                        self.latitude = -self.latitude

                    # Longitude
                    lon_degrees = int(parts[4][:3])
                    lon_minutes = float(parts[4][3:])
                    self.longitude = lon_degrees + lon_minutes / 60.0
                    if parts[5] == "W":
                        self.longitude = -self.longitude

                    # Fix quality
                    self.fix_quality = int(parts[6])

                    # Satellites in use
                    self.satellites_in_use = int(parts[7])
                    
                    # Altitude
                    self.altitude = float(parts[9])

                    return True
                except ValueError:
                    return False
            else:
                return False
        return False
    
    
    def ___parse_nmea_gpvtg(self, nmea_sentence):
        if nmea_sentence.startswith("$GPRMC"):
            parts = nmea_sentence.split(",")
            if len(parts) >= 10 and all(parts[i] for i in [1, 2, 3, 4, 5, 7, 8, 9]): # Check for empty fields
                try:
                    # Time (UTC)
                    self.timestamp = parts[1]

                    # Fix status (A: active, V: void)
                    self.fix_status = parts[2]

                    # Latitude
                    lat_degrees = int(parts[3][:2])
                    lat_minutes = float(parts[3][2:])
                    self.latitude = lat_degrees + lat_minutes / 60.0
                    if parts[4] == "S":
                        self.latitude = -self.latitude

                    # Longitude
                    lon_degrees = int(parts[5][:3])
                    lon_minutes = float(parts[5][3:])
                    self.longitude = lon_degrees + lon_minutes / 60.0
                    if parts[6] == "W":
                        self.longitude = -self.longitude

                    # Speed over ground (knots)
                    self.speed_over_ground_knots = float(parts[7])

                    # Course over ground (degrees)
                    self.course_over_ground = float(parts[8])

                    # Date (DDMMYY)
                    self.date = parts[9] # You might want to store and use the date

                    return True

                except ValueError:
                    return False
            else:
                return False
        return False
    
    def ___parse_nmea_gprmc(self, nmea_sentence):
        if nmea_sentence.startswith("$GPVTG"):
            parts = nmea_sentence.split(",")
            if len(parts) >= 8 and all(parts[i] for i in [1, 5, 7]): # Check for empty fields
                try:
                    # Course over ground (degrees) - True
                    if parts[1]:
                        self.course_over_ground = float(parts[1])
                    # Course over ground (degrees) - Magnetic
                    # if parts[3]:
                    #     self.course_over_ground_magnetic = float(parts[3])
                    # Speed over ground (knots)
                    if parts[5]:
                        self.speed_over_ground_knots = float(parts[5])

                    # Speed over ground (km/h)
                    if parts[7]:
                        self.speed_over_ground_kmh = float(parts[7])
                        
                    return True
                
                except ValueError:
                    return False
            else:
                return False
        return False
    
    def _parse_nmea(self, nmea_sentence):
        """
        Parses an NMEA sentence and extracts relevant data.

        Args:
            nmea_sentence: The NMEA sentence string.

        Returns:
            True if the sentence was parsed successfully, False otherwise.
        """
        if self.gps_mode == self.GPGGA:
            return self.__parse_nmea_gpgga(nmea_sentence)
        elif self.gps_mode == self.GPVTG:
            return self.___parse_nmea_gpvtg(nmea_sentence)
        elif self.gps_mode == self.GPRMC:
            return self.___parse_nmea_gprmc(nmea_sentence)
        return False

    @property
    def latitude(self):
        """
        Get the latitude.

        :return: float: The latitude.
        """
        return self.latitude

    @property
    def longitude(self):
        """
        Get the longitude.

        :return: The longitude (float).
        """
        return self.longitude
    
    @property
    def altitude(self):
        """
        Get the altitude.

        :return: The altitude (float).
        """
        return self.altitude

    @property
    def fix_quality(self):
        """
        Get the fix quality.

        :return: The fix quality (0: invalid, 1: GPS fix, 2: DGPS fix, etc.).
        """
        return self.fix_quality

    @property
    def satellites_in_use(self):
        """
        Get the number of satellites in use.

        :return: The number of satellites in use.
        """
        return self.satellites_in_use

    @property
    def timestamp(self):
        """
        Get the timestamp of the last fix.

        :return: The timestamp in HHMMSS.sss format.
        """
        return self.timestamp
    
    @property
    def fix_status(self):
        """
        Get the fix status from GPRMC sentence.

        :return: 'A' for Active, 'V' for Void (invalid).
        """
        return self.fix_status

    @property
    def speed_knots(self):
        """
        Get the speed over ground in knots.

        :return: The speed over ground in knots (float).
        """
        return self.speed_over_ground_knots

    @property
    def speed_kmh(self):
        """
        Get the speed over ground in km/h.

        :return: The speed over ground in km/h (float).
        """
        return self.speed_over_ground_kmh

    @property
    def course(self):
        """
        Get the course over ground in degrees.

        :return: The course over ground in degrees (float).
        """
        return self.course_over_ground

class Basic:
    class Buzzer:
        def __init__(self):
            self._pin = machine.Pin('P0', machine.Pin.OUT, value=0)

        def on(self):
            self._pin.value(1)

        def off(self):
            self._pin.value(0)

        def beep(self, delay, on=50, off=10):            
            self.on()
            t_on = utime.ticks_ms()
            t_off = 0
            
            while delay:               
                if t_on and utime.ticks_ms() - t_on > on:        
                    self.off()
                    t_on = 0
                    t_off = utime.ticks_ms()
                    
                if t_off and utime.ticks_ms() - t_off > off:
                    t_on = utime.ticks_ms()
                    t_off = 0
                                
                    self.on()  
                    delay -= 1
                        
            self.off()      

    class _I2CtoGPIO:
        """
        MicroPython driver for PCA9535 and PCA9535C 16-bit I2C I/O expanders.
        """
        PCA9535_ADDR = const(0x24)
        
        def __init__(self):
            self._i2c = machine.I2C(1)
            self._i2c.writeto_mem(Basic._I2CtoGPIO.PCA9535_ADDR, 0x06, bytes([0xFF]))
            self._i2c.writeto_mem(Basic._I2CtoGPIO.PCA9535_ADDR, 0x07, bytes([0x00]))
        
        def read(self):
            """
            Reads a register or registers.
            
            :reeturn: A bytes object containing the register value(s).
            """
            return ~self._i2c.readfrom_mem(Basic._I2CtoGPIO.PCA9535_ADDR, 0x00, 1)[0] & 0x03

        def write(self, n):
            """
            Writes to a register or registers.\r
\1 n: The value(s) to write (int or bytes).
            """
            self._i2c.writeto_mem(Basic._I2CtoGPIO.PCA9535_ADDR, 0x03, bytes([0xFF & ~n]))
        
    class _LedIter():
        def __init__(self, leds):
            self._index = 0
            self._leds = leds
            
        def on(self):
            self._leds[self._index-1] = True
        
        def off(self):
            self._leds[self._index-1] = False
            
    class Leds(_I2CtoGPIO):
        def __init__(self):
            super().__init__()
            self._stat = 0x00
            self._count = 8
    
        def __call__(self):
            return tuple([((self._stat >> i) & 0x1) == 1 for i in range(7+1)])
        
        def __getitem__(self, index):
            return ((self._stat >> index) & 0x01) == 1
        
        def __setitem__(self, index, value):
            if value:
                self._stat |= (1 << index)
            else:
                self._stat &= ~(1 << index)
            self.write(self._stat)

        def __iter__(self):
            self._led = Basic._LedIter(self)
            return self
        
        def __next__(self):
            if self._led._index < self._count:
                self._led._index += 1
                return self._led
            else:
                raise StopIteration

        def write(self, n):
            self._stat = n
            super().write(self._stat)
            
        def clear(self):
            self._stat = 0x00
            self.write(self._stat)

    class Buttons(_I2CtoGPIO):
        def __init__(self):
            super().__init__()

        def __getitem__(self, index):
            bits = self.read()
            return (bits & 0x01) if index == 0 else ((bits >> 1) & 0x01) if index == 1 else None
        
        def __call__(self):
            bits = self.read()        
            return bits & 0x01, (bits >> 1) & 0x01

